    include 'mkl_vsl.f90'

    program DGP

    USE random
    USE TOOL
    USE lapack95
    USE MKL_VSL_TYPE
    USE MKL_VSL
    USE omp_lib
    IMPLICIT NONE

    !******************* GENERAL VARIABLES ****************!
    CHARACTER(100)          :: path1='D:/research/2020_exponential_network_formation/simulation/censored/'
    CHARACTER(100)          :: path2='/data/data_40_30/'
    CHARACTER(40)           :: numchr1, numchr2
    INTEGER, PARAMETER      :: NN=40, GG=30,  RR=2000, TT=100, Nthreads=12
    REAL, ALLOCATABLE       :: C(:,:), W(:,:), Wt(:,:), W_new(:,:), Wt_new(:,:), E(:), zero(:), C1(:), Iden(:,:)
    REAL, ALLOCATABLE       :: C_sort(:,:), W_sort(:,:), X_sort(:), Z_sort(:)
    REAL, ALLOCATABLE       :: cov_x(:,:), cov_e(:,:), cov_z(:,:)
    REAL, ALLOCATABLE       :: psi(:,:), X(:), XX(:,:), Z(:), ZZ(:,:), h(:)
    REAL, ALLOCATABLE       :: S(:,:), S1(:,:), S2(:,:),S1L(:,:), S2L(:,:), S_new(:,:)
    REAL, ALLOCATABLE       :: Y(:), Y1(:), Y2(:), Y1s(:), Y2s(:), Y_new(:)
    REAL                    :: gamma(9), beta(2), rho(2), sige, lambda, delta, siga, alpha, rand
    REAL(kind=8)            :: w1, w2, w3, w4, q, pp_w
    INTEGER                 :: clock_max, clock_rate, clock_start, clock_stop, ithread, iblock,ibegin,iend
    INTEGER                 :: brng, method, me, i, j, g, t, flag, ier, r, seedsize, seed(2), errorflag, INFO, seed_mkl
    INTEGER, ALLOCATABLE    :: indx(:)
    LOGICAL, PARAMETER      :: first=.true.
    TYPE (VSL_STREAM_STATE) :: stream
    INTEGER                 :: errcode

    !********************************************************!

    INTERFACE
    SUBROUTINE contraction(ep,W,lambda,Y,NN)
    IMPLICIT NONE
    INTEGER :: NN
    REAL , INTENT(IN)  :: lambda, ep(NN), W(NN,NN)
    REAL, INTENT(OUT)  :: Y(NN)
    END SUBROUTINE contraction
    END INTERFACE


100 format (30F6.2)
200 format (30F6.2)
300 format (1F8.4)


    seedsize=2
    seed(1)=12345
    seed(2)=54321
    CALL RANDOM_SEED
    CALL RANDOM_SEED(SIZE=seedsize)
    CALL RANDOM_SEED(PUT=seed)

    seed_mkl=54321
    brng=VSL_BRNG_MCG31
    method=VSL_RNG_METHOD_GAUSSIANMV_BOXMULLER
    me=VSL_MATRIX_STORAGE_FULL

    !******* Parameters **********!
    gamma=(/-1.7, 0.3, 0.3, 0.2, -0.10, 0.04, 0.30, -0.20,  1.0 /)
    beta=(/ 0.3, 0.1 /)
    rho=(/ 0.3, 0.1  /)
    delta=0.30
    sige=0.5
    lambda=0.05
    siga=0.3
    !***************************!

    call system_clock ( clock_start, clock_rate, clock_max )

    DO t=1, TT

        print*, 'repetition', t
        WRITE(numchr1,*) t

        seed_mkl=seed_mkl+t
        errcode=vslnewstream(stream,brng,seed_mkl)

        Call OMP_SET_NUM_THREADS(Nthreads)
        !$omp parallel default(shared) private(g,C1,C,W,Wt,w_new,wt_new,X,XX,Z,ZZ,i,j,cov_x,cov_e,cov_z,INFO,E,alpha,psi,h, &
        !$omp & r,Y1,Y2,Y1s,Y2s,w1,w2,w3,w4,q, pp_w,rand,Y_new, zero, Iden, errcode, numchr2, indx, &
        !$omp & X_sort, Z_sort, C_sort, W_sort, ithread, iblock, ibegin, iend)

        ithread = OMP_GET_THREAD_NUM()
        iblock = (GG+Nthreads-1)/Nthreads
        ibegin = ithread*iblock+1
        iend = min((ithread+1)*iblock,GG)

        DO g = ibegin,iend
            print*, g

            ALLOCATE(C(NN,NN));      ALLOCATE(C1(NN))
            ALLOCATE(W(NN,NN));      ALLOCATE(Wt(NN,NN))
            ALLOCATE(W_new(NN,NN));  ALLOCATE(Wt_new(NN,NN))
            ALLOCATE(cov_x(NN,NN));  ALLOCATE(cov_e(NN,NN))
            ALLOCATE(cov_z(NN,NN))
            ALLOCATE(X(NN));         ALLOCATE(XX(NN,2))
            ALLOCATE(Z(NN));         ALLOCATE(ZZ(NN,2))
            ALLOCATE(Y1(NN));        ALLOCATE(Y2(NN));
            ALLOCATE(Y1s(NN));       ALLOCATE(Y2s(NN));
            ALLOCATE(zero(NN));      ALLOCATE(E(NN));
            ALLOCATE(Iden(NN,NN))
            ALLOCATE(psi(NN,NN));
            ALLOCATE(h(NN));
            ALLOCATE(Y_new(NN));
            ALLOCATE(X_sort(NN))
            ALLOCATE(indx(NN))
            ALLOCATE(Z_sort(NN))
            ALLOCATE(C_sort(NN,NN))
            ALLOCATE(W_sort(NN,NN))

            DO

                CALL random_number(C1)

                C=0.0
                do i=1,NN
                    do j=1,NN
                        if((C1(i)>=0.7) .and. (C1(j)>=0.7)) C(i,j)=1.0
                        if((C1(i)<=0.3) .and. (C1(j)<=0.3)) C(i,j)=1.0
                    end do
                    C(i,i)=0.0
                end do

                zero=0.0
                Iden=eye(NN)

                cov_x=Iden*4.0
                call spotrf('U', NN, cov_x, NN,INFO)
                errcode=vsrnggaussianmv(method,stream, 1, X, NN, me, zero, cov_x)
                X=X+4.0
                XX(:,1)=X

                cov_z=Iden
                call spotrf('U', NN, cov_z, NN,INFO)
                errcode=vsrnggaussianmv(method,stream, 1, Z, NN, me, zero, cov_z)
                ZZ(:,1)=Z


                cov_e=Iden*sige
                call spotrf('U', NN, cov_e, NN,INFO)
                errcode=vsrnggaussianmv(method,stream, 1, E, NN, me, zero, cov_e)

                alpha=-1.5+random_normal()


                DO i=1, NN
                    DO j=1, NN
                        psi(i,j)=gamma(1)+gamma(2)*C(i,j)-gamma(9)*abs(Z(i)-Z(j))
                    END DO
                END DO


                W=0.0


                XX(:,2)=matmul(W,X)
                ZZ(:,2)=matmul(W,Z)
                h=matmul(XX,beta)+matmul(ZZ,rho)+alpha+E
                CALL contraction(h,W,lambda,Y2,NN)

                W_new=W

                DO r=1, RR
                    DO i=1, NN
                        DO j=1, NN
                            IF (j /= i) THEN
                                W_new(i,j)=1.0-W_new(i,j)
                                XX(:,2)=matmul(W_new,X)
                                ZZ(:,2)=matmul(W_new,Z)
                                h=matmul(XX,beta)+matmul(ZZ,rho)+alpha+E
                                CALL contraction(h,W_new,lambda,Y1,NN)

                                w1=dot_product(W_new(i,:),W_new(j,:))
                                w2=dot_product(W_new(:,i),W_new(:,j))
                                w3=dot_product(W_new(i,:),W_new(:,j))
                                w4=dot_product(W_new(j,:),W_new(:,i))

                                q=psi(i,j) +2*gamma(3)*W_new(j,i)+(2*gamma(4)-gamma(5))*(sum(W_new(i,:))-W_new(i,j)) &
                                    &+3*gamma(5)*(sum(W_new(i,:))-W_new(i,j))**2 &
                                    &+2*gamma(6)*(sum(W_new(:,j))-W_new(i,j)) +gamma(7)*(w1+w2+w3)+gamma(8)*3*w4

                                pp_w=q*(-1)**(1-w_new(i,j))+0.5*delta*(dot_product(Y1,Y1)-dot_product(Y2,Y2))

                                CALL random_number(rand)
                                IF (log(rand)<=pp_w) then
                                    W(i,j)=W_new(i,j)
                                    Y2=Y1
                                END IF
                                W_new(i,j)=W(i,j)
                                XX(:,2)=matmul(W_new,X)
                                ZZ(:,2)=matmul(W_new,Z)
                            END IF
                        END DO
                    END DO
                END DO

                IF ((maxval(sum(W, dim=2))<=20.0) .and. (maxval(sum(W, dim=1))<=20.0)) THEN
                    EXIT
                ELSE
                    print*, 'degenerate'
                END IF
            END DO


            print*, sum(sum(W, dim=2))/NN

            XX(:,2)=matmul(W,X)
            ZZ(:,2)=matmul(W,Z)
            h=matmul(XX,beta)+matmul(ZZ,rho)+alpha+E
            CALL contraction(h,W,lambda,Y_new,NN)

            CALL quick_sort(Y_new,indx)
            DO i=1, NN
                X_sort(i)=X(indx(i))
                Z_sort(i)=Z(indx(i))
                DO j=1, NN
                    C_sort(i,j)=C(indx(i),indx(j))
                    W_sort(i,j)=W(indx(i),indx(j))
                END DO
            END DO

            WRITE(numchr2,*) g

            OPEN (10000+g, file = trim(adjustl(path1))  // trim(adjustl(path2)) // 'R' // trim(adjustl(numchr1)) // '/C/C' // trim(adjustl(numchr2)) // '.txt', ACTION='READWRITE', status='unknown')
            OPEN (20000+g, file = trim(adjustl(path1))  // trim(adjustl(path2)) // 'R' // trim(adjustl(numchr1)) // '/W/W' // trim(adjustl(numchr2)) // '.txt', ACTION='READWRITE', status='unknown')
            OPEN (30000+g, file = trim(adjustl(path1))  // trim(adjustl(path2)) // 'R' // trim(adjustl(numchr1)) // '/X/X' // trim(adjustl(numchr2)) // '.txt', ACTION='READWRITE', status='unknown')
            OPEN (40000+g, file = trim(adjustl(path1))  // trim(adjustl(path2)) // 'R' // trim(adjustl(numchr1)) // '/Y/Y' // trim(adjustl(numchr2)) // '.txt', ACTION='READWRITE', status='unknown')
            OPEN (50000+g, file = trim(adjustl(path1))  // trim(adjustl(path2)) // 'R' // trim(adjustl(numchr1)) // '/alpha/alpha' // trim(adjustl(numchr2)) // '.txt', ACTION='READWRITE', status='unknown')
            OPEN (60000+g, file = trim(adjustl(path1))  // trim(adjustl(path2)) // 'R' // trim(adjustl(numchr1)) // '/Z/Z' // trim(adjustl(numchr2)) // '.txt', ACTION='READWRITE', status='unknown')



            IF(NN==25) THEN
                write(10000+g,'(25f6.2)') ((C_sort(i,j),j=1,NN),i=1,NN)
                write(20000+g,'(25f6.2)') ((W_sort(i,j),j=1,NN),i=1,NN)
            ELSEIF (NN==30) THEN
                write(10000+g,'(30f6.2)') ((C_sort(i,j),j=1,NN),i=1,NN)
                write(20000+g,'(30f6.2)') ((W_sort(i,j),j=1,NN),i=1,NN)
            ELSEIF (NN==40) THEN
                write(10000+g,'(40f6.2)') ((C_sort(i,j),j=1,NN),i=1,NN)
                write(20000+g,'(40f6.2)') ((W_sort(i,j),j=1,NN),i=1,NN)
            END IF

            write(30000+g,300)  X_sort
            write(40000+g,300)  Y_new
            write(50000+g,300)  alpha
            write(60000+g,300)  Z_sort


            close(10000+g)
            close(20000+g)
            close(30000+g)
            close(40000+g)
            close(50000+g)
            close(60000+g)

            DEALLOCATE(C,C1,W,Wt,W_new,Wt_new,cov_x,cov_e,cov_z,X,XX,Z,ZZ,Y1,Y2,Y1s,Y2s,zero,E,Iden,psi,h,Y_new,X_sort,indx,Z_sort,C_sort,W_sort)

        end do
        !$omp end parallel

        call system_clock ( clock_stop, clock_rate, clock_max )
        print*, 'Time:', real ( clock_stop - clock_start ) / real ( clock_rate )


        !***** Deinitialize *****
        errcode=vsldeletestream( stream )

    END DO

    call system_clock ( clock_stop, clock_rate, clock_max )
    print*, 'Time:', real ( clock_stop - clock_start ) / real ( clock_rate )
    pause

    end program DGP



    SUBROUTINE contraction(ep,W,lambda,Y,NN)
    IMPLICIT NONE
    INTEGER            :: NN
    REAL , INTENT(IN)  :: lambda, ep(NN), W(NN,NN)
    REAL, INTENT(OUT)  :: Y(NN)
    REAL :: diff, Y1(NN), Y2(NN), temp_Y(NN,2)

    Y1=ep
    diff=1.0
    Do
        temp_Y(:,1)=lambda*matmul(W,Y1)+ep
        temp_Y(:,2)=0.0
        Y2=maxval(temp_Y,dim=2)
        diff=maxval(abs(Y2-Y1))
        if (diff<=1e-6) exit
        Y1=Y2
    End do
    Y=Y1
    END SUBROUTINE contraction