    include 'mkl_vsl.f90'


    program estimation

    USE random
    USE TOOL
    USE lapack95
    USE MKL_VSL_TYPE
    USE MKL_VSL
    USE omp_lib
    IMPLICIT NONE

    !******************* GENERAL VARIABLES ****************!
    CHARACTER(100)      :: path1='/home/research/2020_exponential_network_formation/simulation/censored'
    CHARACTER(100)      :: path2='/data/data_40_30/'
    CHARACTER(100)      :: path3='/result/censored/latent_only/'
    CHARACTER(40)       :: numchr
    INTEGER, PARAMETER  :: GG=30, TT=50000, MAX=40, RR=1, Nthreads=24, H=10, DD=1, LL=100
    REAL                :: c_1, c_2, acc_1, acc_2, c_3(GG), acc_3(GG), c_0(GG), acc_0(GG), c_4(GG), acc_4(GG), acc_4v, c_5, acc_5, kk
    INTEGER             :: N(GG), NN, missing(GG), flag(GG)
    INTEGER             :: brng, method, me, i, j, g, t, ier, k,  d, r, l, seedsize, seed(2), errorflag, INFO, seed_mkl, t_round
    INTEGER             :: clock_max, clock_rate, clock_start, clock_stop, ithread, iblock,ibegin,iend
    REAL                :: C_raw(MAX,MAX,GG), W_raw(MAX,MAX,GG), X_raw(MAX,GG), Y_raw(MAX,GG), Z_raw(MAX,GG), alpha(GG)
    REAL, ALLOCATABLE   :: C(:,:), W(:,:), Wt(:,:), IPIV(:)
    REAL, ALLOCATABLE   :: Y(:), X(:), XX(:,:)
    LOGICAL, PARAMETER  :: first=.true.
    TYPE (VSL_STREAM_STATE) :: stream
    INTEGER             ::  errcode

    !*** hyperparameters in the prior distribution ***!
    REAL, DIMENSION(2+DD)      :: gamma_0
    REAL, DIMENSION(2+DD,2+DD) :: G_0
    REAL, DIMENSION(4)         :: beta_0
    REAL, DIMENSION(4,4)       :: B_0, B_1, INV_B_0
    REAL                       :: eta_0, A_0
    INTEGER                    :: rho_0

    !*** posterior draws of parameters *******!

    REAL, DIMENSION(2+DD,TT)   :: gamma_T
    REAL, DIMENSION(TT)        :: lambda_T, siga_T, acc_rate1, acc_rate2, acc_rate5
    REAL, DIMENSION(4, TT)     :: beta_T
    REAL, DIMENSION(GG,TT)     :: sige_T, alpha_T, acc_rate0, acc_rate3, acc_rate4

    !******  VARIABLE IN M-H FOR Z_G  *******************************!

    REAL, DIMENSION(MAX,DD,GG):: zz
    REAL, ALLOCATABLE  :: zz_1(:,:), zz_2(:,:), cov_zz(:,:), temp_zz(:)

    !********VARIABLE IN M-H FOR LATENT OTUCOME ***************!

    INTEGER                :: censor(GG)
    REAL, DIMENSION(MAX,GG):: Y_latent
    REAL, ALLOCATABLE :: Y_new(:), Y_old(:), cov_YL(:,:), YL(:)
    REAL, ALLOCATABLE :: W12(:,:), W22(:,:), Wt12(:,:), Wt22(:,:), S(:,:), S22(:,:), S22_new(:,:), S22_old(:,:)


    !********VARIABLE IN M-H FOR GAMMA  & DELTA ***************!

    REAL ::  gamma_1(2+DD), cov_gamma_1(2+DD,2+DD), gamma_2(2+DD)
    REAL, ALLOCATABLE:: psi(:,:), psi_1(:,:), psi_2(:,:), ep(:), zero(:)
    REAL(KIND=8) ::  q_1, q_2, q_g1(GG), pp_g
    INTEGER ::  CODE

    !********VARIABLE IN M-H FOR  LAMBDA  ***************!

    REAL :: lambda_1, lambda_2
    REAL, ALLOCATABLE :: W_old(:,:), W_new(:,:), Wt_old(:,:), Wt_new(:,:)
    REAL :: w1, w2, w3, w4
    REAL, ALLOCATABLE :: S1(:,:), S2(:,:),  S3(:,:), S4(:,:)
    REAL, ALLOCATABLE :: ep_1(:), ep_2(:), ep_1v(:), ep_2v(:)
    REAL, ALLOCATABLE :: V(:,:), V1(:,:), V2(:,:)
    REAL, ALLOCATABLE :: Y1(:),Y2(:),  f(:) , S_new(:,:)
    REAL(KIND=8), ALLOCATABLE ::  pp_w(:,:)
    REAL(KIND=8) :: like_Y1, like_Y2, pp_l(GG), q, rand, pp

    !*************VARIABLE IN M-H FOR BETA**************!

    REAL :: beta_1(4), beta_2(4)
    REAL :: XVX(4,4), XVY(4), INV_B(4,4), INV_B2(4,4), beta_temp(4)
    REAL, ALLOCATABLE ::  XX2(:,:), YY(:)

    !*************VARIABLE IN GIBBS FOR SIGMA_E & SIGMA_EZ & ALPHA_G**************!

    REAL:: alpha_1, alpha_2
    INTEGER :: rho_1, stack

    !********************************************************!

    INTERFACE
    SUBROUTINE contraction(NN,ep,W,lambda,Y)
    IMPLICIT NONE
    INTEGER:: NN
    REAL , INTENT(IN)  :: lambda, ep(:),  W(:,:)
    REAL, INTENT(OUT)  :: Y(:)
    END SUBROUTINE contraction
    END INTERFACE

    seedsize=2
    seed(1)=12345
    seed(2)=54321
    CALL RANDOM_SEED
    CALL RANDOM_SEED(SIZE=seedsize)
    CALL RANDOM_SEED(PUT=seed)

    seed_mkl=12345
    brng=VSL_BRNG_MCG31
    method=VSL_RNG_METHOD_GAUSSIANMV_BOXMULLER
    me=VSL_MATRIX_STORAGE_FULL



    errcode=vslnewstream(stream,brng,seed_mkl)

    DO l=1, LL

        WRITE(numchr,*) l

        CALL networksize(GG, MAX, path1, path2, N)
        CALL define_variable(C_raw, X_raw, W_raw, Y_raw, Z_raw, alpha, GG, l, N, path1, path2, MAX)

        zz=0.0
        beta_0=0.0
        gamma_0=0.0
        eta_0=0.1
        rho_0=2
        B_0=EYE(4)*10.0
        CALL FINDInv(B_0,INV_B_0,4,ier)
        A_0=400.0
        G_0=EYE(2+DD)*10.0


        gamma_T(:,1)=(/-2.3, 0.3, 1.5 /)
        beta_T(:,1)=(/ 0.3, 0.1, 0.3, 0.0 /)
        lambda_T(1)=0.0500
        sige_T(:,1)=0.500
        alpha_T(:,1)=-1.5

        c_0=3e-2
        c_1=1e-6
        c_2=1e-5
        c_3=2.0
        c_4=0.10
        acc_1=0.0;
        acc_2=0.0;acc_3=0.0
        acc_4=0.0;acc_5=0.0
        acc_rate0=0.0;acc_rate1=0.0
        acc_rate2=0.0;acc_rate3=0.0
        acc_rate4=0.0;acc_rate5=0.0

        OPEN (1000, file =  trim(adjustl(path1)) // trim(adjustl(path3)) //  'R' // trim(adjustl(numchr))   // '/gamma.txt', status='unknown')
        OPEN (2000, file =  trim(adjustl(path1)) // trim(adjustl(path3)) //  'R' // trim(adjustl(numchr))   // '/lambda.txt', status='unknown')
        OPEN (3000, file =  trim(adjustl(path1)) // trim(adjustl(path3)) //  'R' // trim(adjustl(numchr))   // '/beta.txt', status='unknown')
        OPEN (4000, file =  trim(adjustl(path1)) // trim(adjustl(path3)) //  'R' // trim(adjustl(numchr))   // '/sigma.txt', status='unknown')
        OPEN (5000, file =  trim(adjustl(path1)) // trim(adjustl(path3)) //  'R' // trim(adjustl(numchr))   // '/alpha.txt', status='unknown')


        Y_latent=0.0
        censor=0
        DO g=1, GG
            NN=N(g)
            Y_latent(1:NN,g)=Y_raw(1:NN,g)
            DO i=1, NN
                IF (Y_latent(i,g)==0) censor(g)=censor(g)+1
            END DO
            IF (censor(g)>0) Y_latent(1:censor(g),g)=-1.5
        END DO

        print*, censor


        DO t=2, TT

            call system_clock ( clock_start, clock_rate, clock_max )

            !*******  M-H STEP TO UPDATE  LATENT Y************!


            Call OMP_SET_NUM_THREADS(Nthreads)

            !$omp parallel default(shared) private(g, NN, W, Wt, X, XX, Y_new, Y_old, i, j,     &
            !$omp & cov_YL, YL, zero, INFO, ep_1, ep_2,                                         &
            !$omp & w12,w22, Wt12,Wt22, S22, r, w1, w2, w3, w4, q, pp_w, rand, zz_1,            &
            !$omp & like_Y1, like_Y2, pp, errcode,                                              &
            !$omp & ithread, iblock, ibegin, iend)

            ithread = OMP_GET_THREAD_NUM()
            iblock = (GG+Nthreads-1)/Nthreads
            ibegin = ithread*iblock+1
            iend = min((ithread+1)*iblock,GG)

            DO g = ibegin,iend
                NN=N(g)
                ALLOCATE(Y_new(NN));       ALLOCATE(Y_old(NN));
                ALLOCATE(X(NN));           ALLOCATE(XX(NN,4))
                ALLOCATE(W(NN,NN));        ALLOCATE(Wt(NN,NN));
                ALLOCATE(ep_1(NN));        ALLOCATE(ep_2(NN));
                ALLOCATE(pp_w(NN,NN))
                ALLOCATE(zz_1(NN,DD))

                Y_new=Y_latent(1:NN,g)
                Y_old=Y_latent(1:NN,g)
                W=W_raw(1:NN,1:NN,g)
                Wt=transpose(W)
                zz_1=zz(1:NN,1:DD,g)
                X=X_raw(1:NN,g)

                XX(:,1)=X
                XX(:,2)=matmul(W,X)
                XX(:,3)=zz_1(:,1)
                XX(:,4)=matmul(W,zz_1(:,1))


                IF (censor(g)>0) THEN

                    DO
                        !!  propose latent Y !!
                        ALLOCATE(cov_YL(censor(g),censor(g)))
                        ALLOCATE(YL(censor(g)))
                        ALLOCATE(zero(censor(g)))
                        zero=Y_latent(1:censor(g),g)
                        cov_YL=eye(censor(g))*c_0(g)
                        CALL spotrf('U',censor(g), cov_YL,censor(g),info)
                        errcode=vsrnggaussianmv(method,stream,1, YL, censor(g) ,me, zero , cov_YL)
                        Y_new(1:censor(g))=YL
                        DEALLOCATE(cov_YL)
                        DEALLOCATE(YL)
                        DEALLOCATE(zero)
                        IF ((maxval(Y_new(1:censor(g)))<0 )) exit ! check if the max YL is still less than zero!
                    END DO

                    ALLOCATE(W12(censor(g),NN-censor(g)))
                    ALLOCATE(W22(NN-censor(g),NN-censor(g)))
                    ALLOCATE(Wt12(censor(g),NN-censor(g)))
                    ALLOCATE(Wt22(NN-censor(g),NN-censor(g)))
                    ALLOCATE(S22(NN-censor(g),NN-censor(g)))

                    W12=W(1:censor(g),censor(g)+1:NN)
                    W22=W(censor(g)+1:NN,censor(g)+1:NN)
                    Wt12=Wt(1:censor(g),censor(g)+1:NN)
                    Wt22=Wt(censor(g)+1:NN,censor(g)+1:NN)
                    S22=eye(NN-censor(g))-lambda_T(t-1)*(W22)

                    !! Define Epsilon !!

                    ep_1(1:censor(g))=Y_new(1:censor(g))-(lambda_T(t-1)*matmul(W12,Y_new(censor(g)+1:NN))+matmul(XX(1:censor(g),:),beta_T(:,t-1))+alpha_T(g,t-1))
                    ep_1(censor(g)+1:NN)=matmul(S22,Y_new(censor(g)+1:NN))-matmul(XX(censor(g)+1:NN,:),beta_T(:,t-1))-alpha_T(g,t-1)

                    ep_2(1:censor(g))=Y_old(1:censor(g))-(lambda_T(t-1)*matmul(W12,Y_old(censor(g)+1:NN))+matmul(XX(1:censor(g),:),beta_T(:,t-1))+alpha_T(g,t-1))
                    ep_2(censor(g)+1:NN)=matmul(S22,Y_old(censor(g)+1:NN))-matmul(XX(censor(g)+1:NN,:),beta_T(:,t-1))-alpha_T(g,t-1)

                    like_Y1=-dot_product(ep_1, ep_1)/(2*sige_T(g,t-1))
                    like_Y2=-dot_product(ep_2, ep_2)/(2*sige_T(g,t-1))

                    pp=like_Y1-like_Y2+logmvnpdf(Y_new(1:censor(g)),eye(censor(g)),censor(g))-logmvnpdf(Y_old(1:censor(g)),eye(censor(g)),censor(g))

                    CALL random_number(rand)
                    IF (log(rand)<=pp) then
                        Y_latent(1:NN,g) = Y_new(1:NN)
                        acc_0(g)=acc_0(g)+1
                    ELSE
                        Y_latent(1:NN,g) = Y_old(1:NN)
                    END IF
                    acc_rate0(g,t)=acc_0(g)/t
                    DEALLOCATE(w12,w22,S22, wt12, wt22)

                    IF((acc_rate0(g,t)<=0.4) .and. (c_0(g)>=1e-6)) c_0(g)=c_0(g)/1.01
                    IF((acc_rate0(g,t)>=0.6) .and. (c_0(g)<=1.0)) c_0(g)=c_0(g)*1.01
                END IF

                DEALLOCATE(Y_new,Y_old,X,XX,W,Wt,ep_1,ep_2,pp_w,zz_1)

            END DO
            !$omp end parallel



            !*******  MH STEP TO UPDATE GAMMA ************!
            
            cov_gamma_1=eye(2+DD)*c_1
            call spotrf('U',2+DD, cov_gamma_1,2+DD,info)
            errcode=vsrnggaussianmv(method,stream,1,gamma_1,2+DD,me,gamma_0, cov_gamma_1)
            gamma_1=gamma_1+gamma_T(:,t-1)
            lambda_1=lambda_T(t-1)+random_normal()*sqrt(c_1)

            IF (t>1000) THEN
                COV_gamma_1=cov(gamma_T(:,1:t-1),2+DD,t-1)*2.38**2/(2+DD)
                call spotrf('U',2+DD, COV_gamma_1, 2+DD, info)
                errcode=vsrnggaussianmv(method,stream,1,gamma_2, 2+DD, me, gamma_0, COV_gamma_1)
                gamma_2=gamma_2+gamma_T(:,t-1)
                gamma_1=0.8*gamma_2+0.2*gamma_1

                lambda_2=random_normal()*sqrt(var(lambda_T(1:t-1),t-1))*2.38+lambda_T(t-1)
                lambda_1=0.8*lambda_2+0.2*lambda_1
            END IF

            gamma_2=gamma_T(:,t-1)

                

            q_g1=0.0
            pp_l=0.0

            Call OMP_SET_NUM_THREADS(Nthreads)

            !$omp parallel default(shared) private(g,NN,C,W,Wt,X,Y,XX, psi, psi_1,psi_2, &
            !$omp & W12,W22,Wt12,Wt22, S22, S22_new, S22_old, ep_1,ep_2, i, j,  pp,    &
            !$omp & INFO, W_old,W_new,Wt_old, Wt_new,  pp_w, zz_1,zz_2, k, kk, &
            !$omp & q,rand,like_Y1,like_Y2,q_1, q_2, &
            !$omp & acc_4v, cov_zz, temp_zz, zero, d, errcode, &
            !$omp & ithread, iblock, ibegin, iend)

            ithread = OMP_GET_THREAD_NUM()
            iblock = (GG+Nthreads-1)/Nthreads
            ibegin = ithread*iblock+1
            iend = min((ithread+1)*iblock,GG)

            DO g = ibegin,iend
                NN=N(g)
                ALLOCATE(C(NN,NN));
                ALLOCATE(W(NN,NN));         ALLOCATE(Wt(NN,NN));
                ALLOCATE(Y(NN));
                ALLOCATE(X(NN));            ALLOCATE(XX(NN,4))
                ALLOCATE(ep_1(NN));         ALLOCATE(ep_2(NN));
                ALLOCATE(psi(NN,NN));
                ALLOCATE(psi_1(NN,NN));     ALLOCATE(psi_2(NN,NN));
                ALLOCATE(w_old(NN,NN));     ALLOCATE(w_new(NN,NN));
                ALLOCATE(Wt_old(NN,NN));    ALLOCATE(Wt_new(NN,NN));
                ALLOCATE(pp_w(NN,NN))
                ALLOCATE(zz_1(NN,DD));      ALLOCATE(zz_2(NN,DD))

                C=C_raw(1:NN,1:NN,g)
                W=W_raw(1:NN,1:NN,g)
                Y=Y_latent(1:NN,g)
                X=X_raw(1:NN,g)
                zz_1=zz(1:NN,1:DD,g)
                zz_2=zz(1:NN,1:DD,g)
                Wt=transpose(W)

                XX(:,1)=X
                XX(:,2)=matmul(W,X)

                DO i=1, NN
                    DO j=1, NN
                        psi(i,j)=gamma_T(1,t-1)+gamma_T(2,t-1)*C(i,j)
                    END DO
                END DO

                ALLOCATE(W12(censor(g),NN-censor(g)))
                ALLOCATE(W22(NN-censor(g),NN-censor(g)))
                ALLOCATE(Wt12(censor(g),NN-censor(g)))
                ALLOCATE(Wt22(NN-censor(g),NN-censor(g)))
                ALLOCATE(S22(NN-censor(g),NN-censor(g)))

                W12=W(1:censor(g),censor(g)+1:NN)
                W22=W(censor(g)+1:NN,censor(g)+1:NN)
                Wt12=Wt(1:censor(g),censor(g)+1:NN)
                Wt22=Wt(censor(g)+1:NN,censor(g)+1:NN)
                S22=eye(NN-censor(g))-lambda_T(t-1)*(W22)

                ! THE DMH ALGORITHM FOR SAMPLING Z !

                kk=NN
                kk=floor(kk/H)
                IF (kk==0) kk=kk+1

                acc_4v=0.0
                DO k=1, kk
                    IF(k==kk) THEN
                        ALLOCATE(cov_zz(NN-(k-1)*H,NN-(k-1)*H))
                        ALLOCATE(temp_zz(NN-(k-1)*H))
                        ALLOCATE(zero(NN-(k-1)*H))
                        zero=0.0
                        cov_zz=EYE(NN-(k-1)*H)*c_4(g)
                        call spotrf('U',NN-(k-1)*H, cov_zz, NN-(k-1)*H, info)
                        DO d=1, DD
                            errcode=vsrnggaussianmv(method,stream, 1, temp_zz, NN-(k-1)*H, me, zero, cov_zz)
                            zz_1((k-1)*H+1:NN,d)=temp_zz+zz_2((k-1)*H+1:NN,d)
                        END DO
                        DEALLOCATE(cov_zz,temp_zz,zero)
                    ELSE
                        ALLOCATE(cov_zz(H,H))
                        ALLOCATE(temp_zz(H))
                        ALLOCATE(zero(H))
                        zero=0.0
                        cov_zz=EYE(H)*c_4(g)
                        call spotrf('U',H, cov_zz, H, info)
                        DO d=1,DD
                            errcode=vsrnggaussianmv(method,stream, 1, temp_zz, H, me, zero, cov_zz)
                            zz_1((k-1)*H+1:k*H,d)=temp_zz+zz_2((k-1)*H+1:k*H,d)
                        END DO
                        DEALLOCATE(cov_zz,temp_zz,zero)
                    END IF

                    pp=0.0
                    psi_1=0.0
                    psi_2=0.0
                    DO i=1,NN
                        DO j=1,NN
                            IF (j/=i) THEN
                                IF (DD==1) THEN
                                    psi_1(i,j)=psi(i,j)-gamma_T(2+DD,t-1)*abs(zz_1(i,1)-zz_1(j,1))
                                    psi_2(i,j)=psi(i,j)-gamma_T(2+DD,t-1)*abs(zz_2(i,1)-zz_2(j,1))
                                ELSEIF (DD==2) THEN
                                    psi_1(i,j)=psi(i,j)-gamma_T(1+DD,t-1)*abs(zz_1(i,1)-zz_1(j,1))-gamma_T(2+DD,t-1)*abs(zz_1(i,2)-zz_1(j,2))
                                    psi_2(i,j)=psi(i,j)-gamma_T(1+DD,t-1)*abs(zz_2(i,1)-zz_2(j,1))-gamma_T(2+DD,t-1)*abs(zz_2(i,2)-zz_2(j,2))
                                END IF
                                q_1=psi_1(i,j)*W(i,j)-log(1+exp(psi_1(i,j)))
                                q_2=psi_2(i,j)*W(i,j)-log(1+exp(psi_2(i,j)))
                                pp=pp+(q_1-q_2)
                            ENDIF
                        END DO
                    END DO

                    XX(:,3)=zz_1(:,1)
                    XX(:,4)=matmul(W,zz_1(:,1))
                    ep_1(1:censor(g))=Y(1:censor(g))-( lambda_T(t-1)*matmul(W12,Y(censor(g)+1:NN))+matmul(XX(1:censor(g),:),beta_T(:,t-1))+alpha_T(g,t-1))
                    ep_1(censor(g)+1:NN)=matmul(S22,Y(censor(g)+1:NN))-matmul(XX(censor(g)+1:NN,:),beta_T(:,t-1))-alpha_T(g,t-1)

                    XX(:,3)=zz_2(:,1)
                    XX(:,4)=matmul(W,zz_2(:,1))
                    ep_2(1:censor(g))=Y(1:censor(g))-( lambda_T(t-1)*matmul(W12,Y(censor(g)+1:NN))+matmul(XX(1:censor(g),:),beta_T(:,t-1))+alpha_T(g,t-1))
                    ep_2(censor(g)+1:NN)=matmul(S22,Y(censor(g)+1:NN))-matmul(XX(censor(g)+1:NN,:),beta_T(:,t-1))-alpha_T(g,t-1)


                    like_Y1=-dot_product(ep_1, ep_1)/(2*sige_T(g,t-1))
                    like_Y2=-dot_product(ep_2, ep_2)/(2*sige_T(g,t-1))


                    IF (k==kk) THEN
                        pp=pp+(like_Y1-like_Y2)+logmvnpdf(reshape(zz_1((k-1)*H+1:NN,:),(/(NN-(k-1)*H)*DD,1/)),eye((NN-(k-1)*H)*DD),(NN-(k-1)*H)*DD)-logmvnpdf(reshape(zz_2((k-1)*H+1:NN,:),(/(NN-(k-1)*H)*DD,1/)),eye((NN-(k-1)*H)*DD),(NN-(k-1)*H)*DD)
                    ELSE
                        pp=pp+(like_Y1-like_Y2)+logmvnpdf(reshape(zz_1((k-1)*H+1:k*H,:),(/H*DD,1/)),eye(H*DD),H*DD)-logmvnpdf(reshape(zz_2((k-1)*H+1:k*H,:),(/H*DD,1/)),eye(H*DD),H*DD)
                    END IF

                    CALL random_number(rand)
                    IF (log(rand)<=pp) THEN
                        zz_2=zz_1
                        acc_4v=acc_4v+1
                    END IF
                    zz_1=zz_2
                END DO
                if (3*acc_4v>=kk) acc_4(g)=acc_4(g)+1
                acc_rate4(g,t)=acc_4(g)/t
                if (acc_rate4(g,t)<0.2)  c_4(g)=c_4(g)/1.01
                if (acc_rate4(g,t)>0.3)  c_4(g)=c_4(g)*1.01
                zz(1:NN,1:DD,g)=zz_1



                !*************************************!
                ! DMH step to update lambda and gamma !
                !*************************************!

                pp=0.0
                psi_1=0.0
                psi_2=0.0
                DO i=1, NN
                    DO j=1, NN
                        IF (j/=i) THEN
                            IF (DD==1) THEN
                                psi_1(i,j)=gamma_1(1)+gamma_1(2)*C(i,j)-gamma_1(2+DD)*abs(zz_1(i,1)-zz_1(j,1))
                                psi_2(i,j)=gamma_2(1)+gamma_2(2)*C(i,j)-gamma_2(2+DD)*abs(zz_1(i,1)-zz_1(j,1))
                            ELSEIF (DD==2) THEN
                                psi_1(i,j)=gamma_1(1)+gamma_1(2)*C(i,j)-gamma_1(1+DD)*abs(zz_1(i,1)-zz_1(j,1))-gamma_1(2+DD)*abs(zz_1(i,2)-zz_1(j,2))
                                psi_2(i,j)=gamma_2(1)+gamma_2(2)*C(i,j)-gamma_2(1+DD)*abs(zz_1(i,1)-zz_1(j,1))-gamma_2(2+DD)*abs(zz_1(i,2)-zz_1(j,2))
                            ENDIF
                            q_1=psi_1(i,j)*W(i,j)-log(1+exp(psi_1(i,j)))
                            q_2=psi_2(i,j)*W(i,j)-log(1+exp(psi_2(i,j)))
                            pp=pp+(q_1-q_2)
                        END IF
                    END DO
                END DO

                XX(:,1)=X
                XX(:,2)=matmul(W,X)
                XX(:,3)=zz_1(:,1)
                XX(:,4)=matmul(W,zz_1(:,1))

                ALLOCATE(S22_new(NN-censor(g),NN-censor(g)))
                ALLOCATE(S22_old(NN-censor(g),NN-censor(g)))

                S22_new=eye(NN-censor(g))-lambda_1*(W22)
                S22_old=eye(NN-censor(g))-lambda_T(t-1)*(W22)

                !! Define Epsilon !!
                ep_1(1:censor(g))=Y(1:censor(g))-(      lambda_1*matmul(W12,Y(censor(g)+1:NN))+matmul(XX(1:censor(g),:),beta_T(:,t-1))+alpha_T(g,t-1))
                ep_1(censor(g)+1:NN)=matmul(S22_new,Y(censor(g)+1:NN))-matmul(XX(censor(g)+1:NN,:),beta_T(:,t-1))-alpha_T(g,t-1)

                ep_2(1:censor(g))=Y(1:censor(g))-( lambda_T(t-1)*matmul(W12,Y(censor(g)+1:NN))+matmul(XX(1:censor(g),:),beta_T(:,t-1))+alpha_T(g,t-1))
                ep_2(censor(g)+1:NN)=matmul(S22_old,Y(censor(g)+1:NN))-matmul(XX(censor(g)+1:NN,:),beta_T(:,t-1))-alpha_T(g,t-1)


                like_Y1=log(FindDet(S22_new, NN-censor(g)))-dot_product(ep_1, ep_1)/(2*sige_T(g,t-1))
                like_Y2=log(FindDet(S22_old, NN-censor(g)))-dot_product(ep_2, ep_2)/(2*sige_T(g,t-1))


                q_g1(g)=pp
                pp_l(g)=(like_Y1-like_Y2)

                DEALLOCATE(C,W,Wt,Y,X,XX,ep_1,ep_2,psi,psi_1,psi_2,w_old,w_new,Wt_old,Wt_new,pp_w,zz_1,zz_2, &
                    & W12, W22, Wt12, Wt22, S22, S22_new, S22_old)

            END DO
            !$omp end parallel
            pp_g=sum(pp_l)+sum(q_g1)+logmvnpdf(gamma_1, G_0, 2+DD)-logmvnpdf(gamma_2, G_0, 2+DD)
            CALL random_number(rand)
            IF (log(rand)<=pp_g)  THEN
                gamma_T(:,t)=gamma_1
                lambda_T(t)=lambda_1
                acc_1=acc_1+1.0
            ELSE
                gamma_T(:,t)=gamma_T(:,t-1)
                lambda_T(t)=lambda_T(t-1)
            END IF
            acc_rate1(t)=acc_1/t


            !********************************************************!
            !******************  DMH STEP FOR BETA  *****************!
            !********************************************************!

            
            B_1=eye(4)*c_2
            call spotrf('U',4, B_1,4,info)
            errcode=vsrnggaussianmv(method,stream,1,beta_1,4,me,beta_0, B_1)
            beta_1=beta_1+beta_T(:,t-1)

            IF (t>1000) THEN
                B_1=cov(beta_T(:,1:t-1),4,t-1)*2.38**2/(4)
                call spotrf('U',4, B_1, 4, info)
                errcode=vsrnggaussianmv(method,stream,1, beta_2, 4, me, beta_0, B_1)
                beta_2=beta_2+beta_T(:,t-1)
                beta_1=0.95*beta_2+0.05*beta_1
            END IF
                

            pp_l=0.0

            Call OMP_SET_NUM_THREADS(Nthreads)
            !$omp parallel default(shared) private(g, NN, W,Wt,Y,X,XX,    &
            !$omp & w12, w22, Wt12, Wt22, S22, ep_1, ep_2, i, j,       &
            !$omp &  q,  rand, INFO, like_Y1, like_Y2,  zz_1, &
            !$omp & ithread, iblock, ibegin, iend)

            ithread = OMP_GET_THREAD_NUM()
            iblock = (GG+Nthreads-1)/Nthreads
            ibegin = ithread*iblock+1
            iend = min((ithread+1)*iblock,GG)
            DO g = ibegin,iend

                NN=N(g)
                ALLOCATE(Wt(NN,NN));
                ALLOCATE(W(NN,NN));         ALLOCATE(Y(NN))
                ALLOCATE(X(NN));            ALLOCATE(XX(NN,4))
                ALLOCATE(ep_1(NN));         ALLOCATE(ep_2(NN))
                ALLOCATE(zz_1(NN,DD))

                W=W_raw(1:NN,1:NN,g)
                Y=Y_latent(1:NN,g)
                X=X_raw(1:NN,g)
                zz_1=zz(1:NN,1:DD,g)
                Wt=transpose(W)

                XX(:,1)=X
                XX(:,2)=matmul(W,X)
                XX(:,3)=zz_1(:,1)
                XX(:,4)=matmul(W,zz_1(:,1))


                ALLOCATE(W12(censor(g),NN-censor(g)))
                ALLOCATE(W22(NN-censor(g),NN-censor(g)))
                ALLOCATE(Wt12(censor(g),NN-censor(g)))
                ALLOCATE(Wt22(NN-censor(g),NN-censor(g)))
                ALLOCATE(S22(NN-censor(g),NN-censor(g)))

                W12=W(1:censor(g),censor(g)+1:NN)
                W22=W(censor(g)+1:NN,censor(g)+1:NN)
                Wt12=Wt(1:censor(g),censor(g)+1:NN)
                Wt22=Wt(censor(g)+1:NN,censor(g)+1:NN)
                S22=eye(NN-censor(g))-lambda_T(t)*(W22)

                !! Define Epsilon !!
                ep_1(1:censor(g))=Y(1:censor(g))-( lambda_T(t)*matmul(W12,Y(censor(g)+1:NN))+matmul(XX(1:censor(g),:),beta_1)+alpha_T(g,t-1))
                ep_1(censor(g)+1:NN)=matmul(S22,Y(censor(g)+1:NN))-matmul(XX(censor(g)+1:NN,:),beta_1)-alpha_T(g,t-1)

                ep_2(1:censor(g))=Y(1:censor(g))-( lambda_T(t)*matmul(W12,Y(censor(g)+1:NN))+matmul(XX(1:censor(g),:),beta_T(:,t-1))+alpha_T(g,t-1))
                ep_2(censor(g)+1:NN)=matmul(S22,Y(censor(g)+1:NN))-matmul(XX(censor(g)+1:NN,:),beta_T(:,t-1))-alpha_T(g,t-1)

                like_Y1=-dot_product(ep_1, ep_1)/(2*sige_T(g,t-1))
                like_Y2=-dot_product(ep_2, ep_2)/(2*sige_T(g,t-1))

                pp_l(g)=(like_Y1-like_Y2)

                DEALLOCATE(W,Wt,Y,X,XX,ep_1,ep_2,zz_1,w12,w22,S22, wt12, wt22)

            END DO
            !$omp end parallel

            pp_g=sum(pp_l)+logmvnpdf(beta_1, B_0, 4)-logmvnpdf(beta_T(:,t-1), B_0, 4)
            CALL random_number(rand)
            IF (log(rand)<=pp_g)  THEN
                beta_T(:,t)=beta_1
                acc_2=acc_2+1.0
            ELSE
                beta_T(:,t)=beta_T(:,t-1)
            END IF
            acc_rate2(t)=acc_2/t


            !******************GIBBS STEP FOR SIGMA_E  *****************!

            DO g=1,GG
                NN=N(g)
                ALLOCATE(W(NN,NN));
                ALLOCATE(Wt(NN,NN))
                ALLOCATE(Y(NN))
                ALLOCATE(X(NN));            ALLOCATE(XX(NN,4))
                ALLOCATE(ep(NN))
                ALLOCATE(zz_1(NN,DD))


                Y=Y_latent(1:NN,g)
                W=W_raw(1:NN,1:NN,g)
                Wt=transpose(W)
                X=X_raw(1:NN,g)
                zz_1=zz(1:NN,1:DD,g)

                XX(:,1)=X
                XX(:,2)=matmul(W,X)
                XX(:,3)=zz_1(:,1)
                XX(:,4)=matmul(W,zz_1(:,1))

                ALLOCATE(W12(censor(g),NN-censor(g)))
                ALLOCATE(W22(NN-censor(g),NN-censor(g)))
                ALLOCATE(Wt12(censor(g),NN-censor(g)))
                ALLOCATE(Wt22(NN-censor(g),NN-censor(g)))
                ALLOCATE(S22(NN-censor(g),NN-censor(g)))

                W12=W(1:censor(g),censor(g)+1:NN)
                W22=W(censor(g)+1:NN,censor(g)+1:NN)
                Wt12=Wt(1:censor(g),censor(g)+1:NN)
                Wt22=Wt(censor(g)+1:NN,censor(g)+1:NN)
                S22=eye(NN-censor(g))-lambda_T(t)*(W22)


                !! Define Epsilon !!
                ep(1:censor(g))=Y(1:censor(g))-( lambda_T(t)*matmul(W12,Y(censor(g)+1:NN))+matmul(XX(1:censor(g),:),beta_T(:,t))+alpha_T(g,t-1))
                ep(censor(g)+1:NN)=matmul(S22,Y(censor(g)+1:NN))-matmul(XX(censor(g)+1:NN,:),beta_T(:,t))-alpha_T(g,t-1)

                sige_T(g,t)=(dot_product(ep,ep)+eta_0)/random_chisq(NN+rho_0, first)

                DEALLOCATE(W12,Wt12,W22,Wt22,S22,W,Wt,X,XX,Y,ep,zz_1)
            END DO



            !!*************GIBBS STEP FOR ALPHA_G*********************!


            Call OMP_SET_NUM_THREADS(Nthreads)
            !$omp parallel default(shared) private(g, NN, W, Wt, Y,X,XX, alpha_1, alpha_2, &
            !$omp & w12, wt12, w22, wt22, S22, ep_1, ep_2, i, j, pp, rand, zz_1, &
            !$omp & INFO, like_Y1, like_Y2 , &
            !$omp & ithread, iblock, ibegin, iend)

            ithread = OMP_GET_THREAD_NUM()
            iblock = (GG+Nthreads-1)/Nthreads
            ibegin = ithread*iblock+1
            iend = min((ithread+1)*iblock,GG)

            DO g = ibegin,iend

                NN=N(g)
                ALLOCATE(W(NN,NN));         ALLOCATE(Wt(NN,NN));
                ALLOCATE(Y(NN))
                ALLOCATE(X(NN));            ALLOCATE(XX(NN,4))
                ALLOCATE(ep_1(NN));         ALLOCATE(ep_2(NN))
                ALLOCATE(zz_1(NN,DD))


                alpha_1=alpha_T(g,t-1)+random_normal()*sqrt(c_3(g))
                IF (t>100) THEN
                    alpha_2=random_normal()*sqrt(var(alpha_T(g,1:t-1),t-1))*2.38+alpha_T(g,t-1)
                    alpha_1=0.95*alpha_2+0.05*alpha_1
                END IF

                W=W_raw(1:NN,1:NN,g)
                Wt=transpose(W)
                Y=Y_latent(1:NN,g)
                X=X_raw(1:NN,g)
                zz_1=zz(1:NN,1:DD,g)

                XX(:,1)=X
                XX(:,2)=matmul(W,X)
                XX(:,3)=zz_1(:,1)
                XX(:,4)=matmul(W,zz_1(:,1))

                ALLOCATE(W12(censor(g),NN-censor(g)))
                ALLOCATE(W22(NN-censor(g),NN-censor(g)))
                ALLOCATE(Wt12(censor(g),NN-censor(g)))
                ALLOCATE(Wt22(NN-censor(g),NN-censor(g)))
                ALLOCATE(S22(NN-censor(g),NN-censor(g)))

                W12=W(1:censor(g),censor(g)+1:NN)
                W22=W(censor(g)+1:NN,censor(g)+1:NN)
                Wt12=Wt(1:censor(g),censor(g)+1:NN)
                Wt22=Wt(censor(g)+1:NN,censor(g)+1:NN)
                S22=eye(NN-censor(g))-lambda_T(t)*(W22)

                !! Define Epsilon !!
                ep_1(1:censor(g))=Y(1:censor(g))-(lambda_T(t)*matmul(W12,Y(censor(g)+1:NN))+matmul(XX(1:censor(g),:),beta_T(:,t))+alpha_1)
                ep_1(censor(g)+1:NN)=matmul(S22,Y(censor(g)+1:NN))-matmul(XX(censor(g)+1:NN,:),beta_T(:,t))-alpha_1

                ep_2(1:censor(g))=Y(1:censor(g))-(lambda_T(t)*matmul(W12,Y(censor(g)+1:NN))+matmul(XX(1:censor(g),:),beta_T(:,t))+alpha_T(g,t-1))
                ep_2(censor(g)+1:NN)=matmul(S22,Y(censor(g)+1:NN))-matmul(XX(censor(g)+1:NN,:),beta_T(:,t))-alpha_T(g,t-1)

                like_Y1=-dot_product(ep_1, ep_1)/(2*sige_T(g,t))
                like_Y2=-dot_product(ep_2, ep_2)/(2*sige_T(g,t))


                pp=like_Y1-like_Y2+((alpha_T(g,t-1)**2-alpha_1**2)/(2*A_0))

                CALL random_number(rand)
                IF (log(rand)<=pp)  THEN
                    alpha_T(g,t)=alpha_1
                    acc_3(g)=acc_3(g)+1.0
                ELSE
                    alpha_T(g,t)=alpha_T(g,t-1)
                END IF
                acc_rate3(g,t)=acc_3(g)/t
                !IF((acc_rate3(g,t)<=0.4) .and. (c_3(g)>=1e-6)) c_3(g)=c_3(g)/1.01
                !IF((acc_rate3(g,t)>=0.6) .and. (c_3(g)<=2)) c_3(g)=c_3(g)*1.01

                DEALLOCATE(W,Wt,Y,X,XX,ep_1,ep_2,zz_1,w12,w22,S22, wt12, wt22)
            END DO
            !$omp end parallel

            call system_clock ( clock_stop, clock_rate, clock_max )
            t_round=t/100
            IF (t-t_round*100==0) THEN
                WRITE(*,*)   'T=', t
                WRITE(*,10) 'Time:', real ( clock_stop - clock_start ) / real ( clock_rate )
                WRITE(*,'(a6, 3f7.4)') 'GAMMA:', gamma_T(:,t)
                WRITE(*,10) 'LAMBDA:', lambda_T(t)
                WRITE(*,10) 'BETA:',   beta_T(:,t)
                WRITE(*,10) 'SIGE:',   sige_T(1,t), sige_T(11,t)
                WRITE(*,10) 'acc_1:',  acc_rate1(t)
                WRITE(*,10) 'acc_2:',  acc_rate2(t)
                WRITE(*,10) 'acc_3:',  acc_rate3(1,t)
                WRITE(*,10) 'acc_4:',  acc_rate4(1,t)
                WRITE(*,10) 'c_4',     c_4(1), c_4(11)
                WRITE(*,10) 'zz',      zz(1:5,1,1)
            END IF
            WRITE(1000,'(14f10.4)')  gamma_T(:,t)
            WRITE(2000,'(f10.4)')    lambda_T(t)
            WRITE(3000,'(30f10.4)')  beta_T(:,t)
            WRITE(4000,'(47f10.4)')  sige_T(:,t)
            WRITE(5000,'(47f10.4)')  alpha_T(:,t)

        END DO
    END DO

10  format(a6,5f7.4)
30  format(a6,8f7.4)
40  format(a6,6f10.4)
50  format(a6,5f10.4)
60  format(a6,f10.4)
70  format(30f10.4)


    end program estimation

    SUBROUTINE contraction(NN,ep,W,lambda,Y)
    IMPLICIT NONE
    INTEGER:: NN
    REAL , INTENT(IN)  :: lambda, ep(:),  W(:,:)
    REAL, INTENT(OUT)  :: Y(:)

    ! local variable !
    REAL, ALLOCATABLE :: Y1(:), Y2(:), temp_Y(:,:)
    REAL :: diff
    INTEGER:: count

    ALLOCATE(Y1(NN))
    ALLOCATE(Y2(NN))
    ALLOCATE(temp_Y(NN,2))

    Y1=ep
    diff=1.0
    count=0
    Do
        temp_Y(:,1)=lambda*matmul(W,Y1)+ep
        temp_Y(:,2)=0.0
        Y2=maxval(temp_Y,dim=2)
        diff=maxval(abs(Y2-Y1))
        if (diff<=1e-6) exit
        Y1=Y2
        count=count+1
        if (count>=200) then
            !print*, 'convergence fail'
            exit
        endif
    End do
    Y=Y2

    DEALLOCATE(Y1)
    DEALLOCATE(Y2)
    DEALLOCATE(temp_Y)
    END SUBROUTINE contraction