    include 'mkl_vsl.f90'


    program estimation

    USE random
    USE TOOL
    USE lapack95
    USE MKL_VSL_TYPE
    USE MKL_VSL
    USE omp_lib
    IMPLICIT NONE


    !******************* GENERAL VARIABLES ****************!
    CHARACTER(100)      :: path1='/home/research/2020_exponential_network_formation/simulation/censored'
    CHARACTER(100)      :: path2='/data/data_40_30/'
    CHARACTER(100)      :: path3='/result/censored/no_latent/'
    CHARACTER(40)       :: numchr
    INTEGER, PARAMETER  :: GG=30, TT=50000, MAX=40, RR=1, Nthreads=24, H=10, DD=1, LL=100
    REAL                :: c_1, c_2, acc_1, acc_2, c_3(GG), acc_3(GG), c_0(GG), acc_0(GG), c_4(GG), acc_4(GG), acc_4v, c_5, acc_5, kk
    INTEGER             :: N(GG), NN, missing(GG), flag(GG)
    INTEGER             :: brng, method, me, i, j, g, t, ier, k,  d, r, l, seedsize, seed(2), errorflag, INFO, seed_mkl, t_round
    INTEGER             :: clock_max, clock_rate, clock_start, clock_stop, ithread, iblock,ibegin,iend
    REAL                :: C_raw(MAX,MAX,GG), W_raw(MAX,MAX,GG), X_raw(MAX,GG), Y_raw(MAX,GG), Z_raw(MAX,GG), alpha(GG)
    REAL, ALLOCATABLE   :: C(:,:), W(:,:), Wt(:,:), IPIV(:)
    REAL, ALLOCATABLE   :: Y(:), X(:), XX(:,:)
    LOGICAL, PARAMETER  :: first=.true.
    TYPE (VSL_STREAM_STATE) :: stream
    INTEGER             ::  errcode

    !*** hyperparameters in the prior distribution ***!
    REAL, DIMENSION(9)      :: gamma_0
    REAL, DIMENSION(9,9)    :: G_0
    REAL, DIMENSION(2)         :: beta_0
    REAL, DIMENSION(2,2)       :: B_0, B_1, INV_B_0
    REAL                       :: eta_0, A_0
    INTEGER                    :: rho_0

    !*** posterior draws of parameters *******!

    REAL, DIMENSION(9,TT)   :: gamma_T
    REAL, DIMENSION(TT)        :: lambda_T, siga_T, acc_rate1, acc_rate2, acc_rate5
    REAL, DIMENSION(2, TT)     :: beta_T
    REAL, DIMENSION(GG,TT)     :: sige_T, alpha_T, acc_rate0, acc_rate3, acc_rate4


    !********VARIABLE IN M-H FOR LATENT OTUCOME ***************!

    INTEGER                :: censor(GG)
    REAL, DIMENSION(MAX,GG):: Y_latent
    REAL, ALLOCATABLE :: Y_new(:), Y_old(:), cov_YL(:,:), YL(:)
    REAL, ALLOCATABLE :: W12(:,:), W22(:,:), Wt12(:,:), Wt22(:,:), S(:,:), S22(:,:), S22_new(:,:), S22_old(:,:)


    !********VARIABLE IN M-H FOR GAMMA  & DELTA ***************!

    REAL ::  gamma_1(9), cov_gamma_1(9,9), gamma_2(9)
    REAL, ALLOCATABLE:: psi(:,:), psi_1(:,:), psi_2(:,:), ep(:), zero(:)
    REAL(KIND=8) ::  q_1, q_2, q_g1(GG), pp_g
    INTEGER ::  CODE

    !********VARIABLE IN M-H FOR  LAMBDA  ***************!

    REAL :: lambda_1, lambda_2
    REAL, ALLOCATABLE :: W_old(:,:), W_new(:,:), Wt_old(:,:), Wt_new(:,:)
    REAL :: w1, w2, w3, w4
    REAL, ALLOCATABLE :: S1(:,:), S2(:,:),  S3(:,:), S4(:,:)
    REAL, ALLOCATABLE :: ep_1(:), ep_2(:), ep_1v(:), ep_2v(:)
    REAL, ALLOCATABLE :: V(:,:), V1(:,:), V2(:,:)
    REAL, ALLOCATABLE :: Y1(:),Y2(:), Y1s(:), Y2s(:), Y3(:),Y4(:), Y3s(:), Y4s(:), h1(:), h2(:), f(:) , S_new(:,:)
    REAL(KIND=8), ALLOCATABLE ::  pp_w(:,:)
    REAL(KIND=8) :: like_Y1, like_Y2, pp_l(GG), q, rand, pp

    !*************VARIABLE IN M-H FOR BETA**************!

    REAL :: beta_1(2), beta_2(2)
    REAL :: XVX(2,2), XVY(2), INV_B(2,2), INV_B2(2,2), beta_temp(2)
    REAL, ALLOCATABLE ::  XX2(:,:), YY(:)

    !*************VARIABLE IN GIBBS FOR SIGMA_E & SIGMA_EZ & ALPHA_G**************!

    REAL:: alpha_1, alpha_2
    INTEGER :: rho_1, stack

    !********************************************************!

    INTERFACE
    SUBROUTINE contraction(NN,ep,W,lambda,Y)
    IMPLICIT NONE
    INTEGER:: NN
    REAL , INTENT(IN)  :: lambda, ep(:),  W(:,:) 
    REAL, INTENT(OUT)  :: Y(:)
    END SUBROUTINE contraction
    END INTERFACE

    seedsize=2
    seed(1)=12345
    seed(2)=54321
    CALL RANDOM_SEED
    CALL RANDOM_SEED(SIZE=seedsize)
    CALL RANDOM_SEED(PUT=seed)

    seed_mkl=12345
    brng=VSL_BRNG_MCG31
    method=VSL_RNG_METHOD_GAUSSIANMV_BOXMULLER
    me=VSL_MATRIX_STORAGE_FULL



    errcode=vslnewstream(stream,brng,seed_mkl)

    DO l=1, LL

        WRITE(numchr,*) l

        CALL networksize(GG, MAX, path1, path2, N)
        CALL define_variable(C_raw, X_raw, W_raw, Y_raw, Z_raw, alpha, GG, l, N, path1, path2, MAX)

        beta_0=0.0
        gamma_0=0.0
        eta_0=0.1
        rho_0=2
        B_0=EYE(2)*10.0
        CALL FINDInv(B_0,INV_B_0,2,ier)
        A_0=400.0
        G_0=EYE(9)*10.0


        gamma_T(:,1)=(/-2.3, 0.3, 0.3, 0.2, -0.1, 0.03, 0.30, -0.20, 0.3 /)
        beta_T(:,1)=(/ 0.3, 0.1 /)
        lambda_T(1)=0.0700
        sige_T(:,1)=0.500
        alpha_T(:,1)=0.5

        c_0=3e-2
        c_1=1e-6
        c_2=1e-4
        c_3=2.0
        c_4=0.10
        acc_1=0.0;
        acc_2=0.0;acc_3=0.0
        acc_4=0.0;acc_5=0.0
        acc_rate0=0.0;acc_rate1=0.0
        acc_rate2=0.0;acc_rate3=0.0
        acc_rate4=0.0;acc_rate5=0.0

        OPEN (1000, file =  trim(adjustl(path1)) // trim(adjustl(path3)) //  'R' // trim(adjustl(numchr))   // '/gamma.txt', status='unknown')
        OPEN (2000, file =  trim(adjustl(path1)) // trim(adjustl(path3)) //  'R' // trim(adjustl(numchr))   // '/lambda.txt', status='unknown')
        OPEN (3000, file =  trim(adjustl(path1)) // trim(adjustl(path3)) //  'R' // trim(adjustl(numchr))   // '/beta.txt', status='unknown')
        OPEN (4000, file =  trim(adjustl(path1)) // trim(adjustl(path3)) //  'R' // trim(adjustl(numchr))   // '/sigma.txt', status='unknown')
        OPEN (5000, file =  trim(adjustl(path1)) // trim(adjustl(path3)) //  'R' // trim(adjustl(numchr))   // '/alpha.txt', status='unknown')


        Y_latent=0.0
        censor=0
        DO g=1, GG
            NN=N(g)
            Y_latent(1:NN,g)=Y_raw(1:NN,g)
            DO i=1, NN
                IF (Y_latent(i,g)==0) censor(g)=censor(g)+1
            END DO
            IF (censor(g)>0) Y_latent(1:censor(g),g)=-1.5
        END DO

        print*, censor
        pause


        DO t=2, TT

            call system_clock ( clock_start, clock_rate, clock_max )

            !*******  M-H STEP TO UPDATE  LATENT Y************!


            Call OMP_SET_NUM_THREADS(Nthreads)

            !$omp parallel default(shared) private(g, NN, C, W, Wt, X, XX, Y_new, Y_old, i, j,  &
            !$omp & cov_YL, YL, zero, INFO, ep_1, ep_2, psi, W_new, W_old, Wt_new, Wt_old,      &
            !$omp & w12,w22, Wt12,Wt22, S22, r, w1, w2, w3, w4, q, pp_w, rand,  h1, h2,         &
            !$omp & Y1, Y2, Y1s, Y2s, like_Y1, like_Y2, pp, errcode,                            &
            !$omp & ithread, iblock, ibegin, iend)

            ithread = OMP_GET_THREAD_NUM()
            iblock = (GG+Nthreads-1)/Nthreads
            ibegin = ithread*iblock+1
            iend = min((ithread+1)*iblock,GG)

            DO g = ibegin,iend
                NN=N(g)
                ALLOCATE(Y_new(NN));       ALLOCATE(Y_old(NN));
                ALLOCATE(C(NN,NN))
                ALLOCATE(X(NN));           ALLOCATE(XX(NN,2))
                ALLOCATE(W(NN,NN));        ALLOCATE(Wt(NN,NN));
                ALLOCATE(ep_1(NN));        ALLOCATE(ep_2(NN));
                ALLOCATE(pp_w(NN,NN))
                ALLOCATE(psi(NN,NN))
                ALLOCATE(W_new(NN,NN));    ALLOCATE(W_old(NN,NN))
                ALLOCATE(Wt_new(NN,NN));   ALLOCATE(Wt_old(NN,NN))
                ALLOCATE(h1(NN));          ALLOCATE(h2(NN))
                ALLOCATE(Y1(NN));          ALLOCATE(Y2(NN))
                ALLOCATE(Y1s(NN));         ALLOCATE(Y2s(NN))

                Y_new=Y_latent(1:NN,g)
                Y_old=Y_latent(1:NN,g)
                C=C_raw(1:NN,1:NN,g)
                W=W_raw(1:NN,1:NN,g)
                Wt=transpose(W)
                X=X_raw(1:NN,g)

                XX(:,1)=X
                XX(:,2)=matmul(W,X)


                IF (censor(g)>0) THEN

                    DO
                        !!  propose latent Y !!
                        ALLOCATE(cov_YL(censor(g),censor(g)))
                        ALLOCATE(YL(censor(g)))
                        ALLOCATE(zero(censor(g)))
                        zero=Y_latent(1:censor(g),g)
                        cov_YL=eye(censor(g))*c_0(g)
                        CALL spotrf('U',censor(g), cov_YL,censor(g),info)
                        errcode=vsrnggaussianmv(method,stream,1, YL, censor(g) ,me, zero , cov_YL)
                        Y_new(1:censor(g))=YL
                        DEALLOCATE(cov_YL)
                        DEALLOCATE(YL)
                        DEALLOCATE(zero)
                        IF ((maxval(Y_new(1:censor(g)))<0 )) exit ! check if the max YL is still less than zero!
                    END DO

                    ALLOCATE(W12(censor(g),NN-censor(g)))
                    ALLOCATE(W22(NN-censor(g),NN-censor(g)))
                    ALLOCATE(Wt12(censor(g),NN-censor(g)))
                    ALLOCATE(Wt22(NN-censor(g),NN-censor(g)))
                    ALLOCATE(S22(NN-censor(g),NN-censor(g)))

                    W12=W(1:censor(g),censor(g)+1:NN)
                    W22=W(censor(g)+1:NN,censor(g)+1:NN)
                    Wt12=Wt(1:censor(g),censor(g)+1:NN)
                    Wt22=Wt(censor(g)+1:NN,censor(g)+1:NN)
                    S22=eye(NN-censor(g))-lambda_T(t-1)*(W22)

                    !! Define Epsilon !!

                    ep_1(1:censor(g))=Y_new(1:censor(g))-(lambda_T(t-1)*matmul(W12,Y_new(censor(g)+1:NN))+matmul(XX(1:censor(g),:),beta_T(:,t-1))+alpha_T(g,t-1))
                    ep_1(censor(g)+1:NN)=matmul(S22,Y_new(censor(g)+1:NN))-matmul(XX(censor(g)+1:NN,:),beta_T(:,t-1))-alpha_T(g,t-1)

                    ep_2(1:censor(g))=Y_old(1:censor(g))-(lambda_T(t-1)*matmul(W12,Y_old(censor(g)+1:NN))+matmul(XX(1:censor(g),:),beta_T(:,t-1))+alpha_T(g,t-1))
                    ep_2(censor(g)+1:NN)=matmul(S22,Y_old(censor(g)+1:NN))-matmul(XX(censor(g)+1:NN,:),beta_T(:,t-1))-alpha_T(g,t-1)

                    DO i=1, NN
                        DO j=1, NN
                            psi(i,j)=gamma_T(1,t-1)+gamma_T(2,t-1)*C(i,j)
                        END DO
                    END DO

                    W_new=W
                    W_old=W
                    
                    XX(:,2)=matmul(w_old,X)
                    h1=ep_1+matmul(XX,beta_T(:,t-1))+alpha_T(g,t-1)
                    CALL contraction(NN,h1,W_old,lambda_T(t-1),Y2)

                    DO r=1, RR
                        DO i=1, NN
                            DO j=1, NN
                                IF (j /= i) THEN
                                    w_new(i,j)=1-w_new(i,j)
                                    XX(:,2)=matmul(w_new,X)
                                    h1=ep_1+matmul(XX,beta_T(:,t-1))+alpha_T(g,t-1)
                                    CALL contraction(NN,h1,W_new,lambda_T(t-1),Y1)

                                    w1=dot_product(w_new(i,:),w_new(j,:))
                                    w2=dot_product(w_new(:,i),w_new(:,j))
                                    w3=dot_product(w_new(i,:),w_new(:,j))
                                    w4=dot_product(w_new(j,:),w_new(:,i))

                                    q=psi(i,j)+2*gamma_T(3,t-1)*w_new(j,i)+(2*gamma_T(4,t-1)-gamma_T(5,t-1))*(sum(w_new(i,:))-w_new(i,j)) &
                                        &+3*gamma_T(5,t-1)*(sum(w_new(i,:))-w_new(i,j))**2 &
                                        &+2*gamma_T(6,t-1)*(sum(w_new(:,j))-w_new(i,j)) +gamma_T(7,t-1)*(w1+w2+w3)+gamma_T(8,t-1)*3*w4

                                    pp_w(i,j)=q*(-1)**(1-w_new(i,j))+0.5*gamma_T(9,t-1)*(dot_product(Y1,Y1)-dot_product(Y2,Y2))

                                    CALL random_number(rand)
                                    IF (log(rand)<=pp_w(i,j)) then
                                        W_old(i,j)=W_new(i,j)
                                        Y2=Y1
                                    END IF
                                    W_new(i,j)=W_old(i,j)
                                    XX(:,2)=matmul(w_new,X)
                                END IF
                            END DO
                        END DO
                    END DO


                    like_Y1=-dot_product(ep_1, ep_1)/(2*sige_T(g,t-1))
                    like_Y2=-dot_product(ep_2, ep_2)/(2*sige_T(g,t-1))


                    XX(:,2)=matmul(w_new,X)
                    h1=ep_1+matmul(XX,beta_T(:,t-1))+alpha_T(g,t-1)
                    h2=ep_2+matmul(XX,beta_T(:,t-1))+alpha_T(g,t-1)
                    CALL contraction(NN,h2,W_new,lambda_T(t-1),Y1)
                    CALL contraction(NN,h1,W_new,lambda_T(t-1),Y2)

                    pp=like_Y1-like_Y2 + 0.5*gamma_T(9,t-1)*(dot_product(Y1,Y1)-dot_product(Y2,Y2))    &
                        & +logmvnpdf(Y_new(1:censor(g)),eye(censor(g)),censor(g))-logmvnpdf(Y_old(1:censor(g)),eye(censor(g)),censor(g))

                    CALL random_number(rand)
                    IF (log(rand)<=pp) then
                        Y_latent(1:NN,g) = Y_new(1:NN)
                        acc_0(g)=acc_0(g)+1
                    ELSE
                        Y_latent(1:NN,g) = Y_old(1:NN)
                    END IF
                    acc_rate0(g,t)=acc_0(g)/t
                    DEALLOCATE(w12,w22,S22, wt12, wt22)

                    IF((acc_rate0(g,t)<=0.4) .and. (c_0(g)>=1e-6)) c_0(g)=c_0(g)/1.01
                    IF((acc_rate0(g,t)>=0.6) .and. (c_0(g)<=1.0)) c_0(g)=c_0(g)*1.01
                END IF

                DEALLOCATE(Y_new,Y_old,C,X,XX,W,Wt,ep_1,ep_2,pp_w,psi,W_new,W_old,Wt_new,Wt_old,h1,h2,Y1,Y2,Y1s,Y2s)

            END DO
            !$omp end parallel



            !*******  MH STEP TO UPDATE GAMMA ************!

           
            cov_gamma_1=eye(9)*c_1
            call spotrf('U',9, cov_gamma_1,9,info)
            errcode=vsrnggaussianmv(method,stream,1,gamma_1,9,me,gamma_0, cov_gamma_1)
            gamma_1=gamma_1+gamma_T(:,t-1)
            lambda_1=lambda_T(t-1)+random_normal()*sqrt(c_1)

            IF (t>1000) THEN
                COV_gamma_1=cov(gamma_T(:,1:t-1),9,t-1)*2.38**2/(9)
                call spotrf('U',9, COV_gamma_1, 9, info)
                errcode=vsrnggaussianmv(method,stream,1,gamma_2, 9, me, gamma_0, COV_gamma_1)
                gamma_2=gamma_2+gamma_T(:,t-1)
                gamma_1=0.8*gamma_2+0.2*gamma_1

                lambda_2=random_normal()*sqrt(var(lambda_T(1:t-1),t-1))*2.38+lambda_T(t-1)
                lambda_1=0.8*lambda_2+0.2*lambda_1
            END IF

            gamma_2=gamma_T(:,t-1)

               

            q_g1=0.0
            pp_l=0.0

            Call OMP_SET_NUM_THREADS(Nthreads)

            !$omp parallel default(shared) private(g,NN,C,W,Wt,X,Y,XX, psi, psi_1,psi_2, &
            !$omp & W12,W22,Wt12,Wt22, S22, S22_new, S22_old, ep_1,ep_2, i, j, r, pp,    &
            !$omp & INFO, W_old,W_new,Wt_old, Wt_new,  pp_w, h1, h2, Y1, Y2,  Y1s, Y2s, k, kk, &
            !$omp & q, rand, like_Y1,like_Y2,q_1, q_2, w1, w2, w3, w4, Y3, Y4, Y3s, Y4s, &
            !$omp & errcode, &
            !$omp & ithread, iblock, ibegin, iend)

            ithread = OMP_GET_THREAD_NUM()
            iblock = (GG+Nthreads-1)/Nthreads
            ibegin = ithread*iblock+1
            iend = min((ithread+1)*iblock,GG)

            DO g = ibegin,iend
                NN=N(g)
                ALLOCATE(C(NN,NN));
                ALLOCATE(W(NN,NN));         ALLOCATE(Wt(NN,NN));
                ALLOCATE(Y(NN));
                ALLOCATE(X(NN));            ALLOCATE(XX(NN,2))
                ALLOCATE(ep_1(NN));         ALLOCATE(ep_2(NN));
                ALLOCATE(psi(NN,NN));
                ALLOCATE(psi_1(NN,NN));     ALLOCATE(psi_2(NN,NN));
                ALLOCATE(w_old(NN,NN));     ALLOCATE(w_new(NN,NN));
                ALLOCATE(Wt_old(NN,NN));    ALLOCATE(Wt_new(NN,NN));
                ALLOCATE(pp_w(NN,NN))
                ALLOCATE(h1(NN));          ALLOCATE(h2(NN))
                ALLOCATE(Y1(NN));          ALLOCATE(Y2(NN))
                ALLOCATE(Y1s(NN));         ALLOCATE(Y2s(NN))
                ALLOCATE(Y3(NN));          ALLOCATE(Y4(NN))
                ALLOCATE(Y3s(NN));         ALLOCATE(Y4s(NN))


                C=C_raw(1:NN,1:NN,g)
                W=W_raw(1:NN,1:NN,g)
                Y=Y_latent(1:NN,g)
                X=X_raw(1:NN,g)
                Wt=transpose(W)

                XX(:,1)=X
                XX(:,2)=matmul(W,X)

                
                ALLOCATE(W12(censor(g),NN-censor(g)))
                ALLOCATE(W22(NN-censor(g),NN-censor(g)))
                ALLOCATE(Wt12(censor(g),NN-censor(g)))
                ALLOCATE(Wt22(NN-censor(g),NN-censor(g)))
                ALLOCATE(S22(NN-censor(g),NN-censor(g)))

                W12=W(1:censor(g),censor(g)+1:NN)
                W22=W(censor(g)+1:NN,censor(g)+1:NN)
                Wt12=Wt(1:censor(g),censor(g)+1:NN)
                Wt22=Wt(censor(g)+1:NN,censor(g)+1:NN)
                S22=eye(NN-censor(g))-lambda_T(t-1)*(W22)



                !*************************************!
                ! DMH step to update lambda and gamma !
                !*************************************!

                psi_1=0.0
                psi_2=0.0
                DO i=1, NN
                    DO j=1, NN
                        IF (j/=i) THEN
                            psi_1(i,j)=gamma_1(1)+gamma_1(2)*C(i,j)
                            psi_2(i,j)=gamma_2(1)+gamma_2(2)*C(i,j)
                        END IF
                    END DO
                END DO

                XX(:,1)=X
                XX(:,2)=matmul(W,X)

                ALLOCATE(S22_new(NN-censor(g),NN-censor(g)))
                ALLOCATE(S22_old(NN-censor(g),NN-censor(g)))

                S22_new=eye(NN-censor(g))-lambda_1*(W22)
                S22_old=eye(NN-censor(g))-lambda_T(t-1)*(W22)

                !! Define Epsilon !!
                ep_1(1:censor(g))=Y(1:censor(g))-(      lambda_1*matmul(W12,Y(censor(g)+1:NN))+matmul(XX(1:censor(g),:),beta_T(:,t-1))+alpha_T(g,t-1))
                ep_1(censor(g)+1:NN)=matmul(S22_new,Y(censor(g)+1:NN))-matmul(XX(censor(g)+1:NN,:),beta_T(:,t-1))-alpha_T(g,t-1)

                ep_2(1:censor(g))=Y(1:censor(g))-( lambda_T(t-1)*matmul(W12,Y(censor(g)+1:NN))+matmul(XX(1:censor(g),:),beta_T(:,t-1))+alpha_T(g,t-1))
                ep_2(censor(g)+1:NN)=matmul(S22_old,Y(censor(g)+1:NN))-matmul(XX(censor(g)+1:NN,:),beta_T(:,t-1))-alpha_T(g,t-1)

                W_new=W
                W_old=W
                Wt_new=Wt
                Wt_old=Wt

                XX(:,2)=matmul(w_old,X)
                h1=ep_1+matmul(XX,beta_T(:,t-1))+alpha_T(g,t-1)
                CALL contraction(NN,h1,W_old,lambda_1,Y2)

                DO r=1, RR
                    DO i=1, NN
                        DO j=1, NN
                            IF (j /= i) THEN
                                w_new(i,j)=1-w_new(i,j)
                                XX(:,2)=matmul(w_new,X)
                                h1=ep_1+matmul(XX,beta_T(:,t-1))+alpha_T(g,t-1)
                                CALL contraction(NN,h1,W_new,lambda_1,Y1)

                                w1=dot_product(w_new(i,:),w_new(j,:))
                                w2=dot_product(w_new(:,i),w_new(:,j))
                                w3=dot_product(w_new(i,:),w_new(:,j))
                                w4=dot_product(w_new(j,:),w_new(:,i))

                                q=psi_1(i,j)+2*gamma_1(3)*w_new(j,i)+(2*gamma_1(4)-gamma_1(5))*(sum(w_new(i,:))-w_new(i,j)) &
                                    &+3*gamma_1(5)*(sum(w_new(i,:))-w_new(i,j))**2 &
                                    &+2*gamma_1(6)*(sum(w_new(:,j))-w_new(i,j)) +gamma_1(7)*(w1+w2+w3)+gamma_1(8)*3*w4

                                pp_w(i,j)=q*(-1)**(1-w_new(i,j))+0.5*gamma_1(9)*(dot_product(Y1,Y1)-dot_product(Y2,Y2))

                                CALL random_number(rand)
                                IF (log(rand)<=pp_w(i,j)) then
                                    W_old(i,j)=W_new(i,j)
                                    Y2=Y1
                                END IF
                                W_new(i,j)=W_old(i,j)
                                XX(:,2)=matmul(w_new,X)
                            END IF
                        END DO
                    END DO
                END DO


                like_Y1=log(FindDet(S22_new, NN-censor(g)))-dot_product(ep_1, ep_1)/(2*sige_T(g,t-1))
                like_Y2=log(FindDet(S22_old, NN-censor(g)))-dot_product(ep_2, ep_2)/(2*sige_T(g,t-1))


                XX(:,1)=X
                XX(:,2)=matmul(w,X)
                h1=ep_1+matmul(XX,beta_T(:,t-1))+alpha_T(g,t-1)
                h2=ep_2+matmul(XX,beta_T(:,t-1))+alpha_T(g,t-1)
                CALL contraction(NN,h1,W,lambda_1,Y1)
                CALL contraction(NN,h2,W,lambda_T(t-1),Y2)

                XX(:,2)=matmul(w_new,X)
                h1=ep_1+matmul(XX,beta_T(:,t-1))+alpha_T(g,t-1)
                h2=ep_2+matmul(XX,beta_T(:,t-1))+alpha_T(g,t-1)
                CALL contraction(NN,h2,W_new,lambda_T(t-1),Y3)
                CALL contraction(NN,h1,W_new,lambda_1,Y4)

                q_1=0.5*gamma_1(9)*dot_product(Y1,Y1)-0.5*gamma_2(9)*dot_product(Y2,Y2) &
                    &+trace(matmul(psi_1,transpose(W)),NN)- trace(matmul(psi_2,transpose(W)),NN) &
                    &+(gamma_1(3)-gamma_2(3))*trace(matmul(W,W),NN)&
                    &+(gamma_1(4)-gamma_2(4))*(sum(matmul(W,ones(NN,1))*matmul(W,ones(NN,1)))-sum( matmul(W,ones(NN,1)))) &
                    &+(gamma_1(5)-gamma_2(5))*(sum(matmul(W,ones(NN,1))*(matmul(W,ones(NN,1))*matmul(W,ones(NN,1))))&
                    &-2*sum(matmul(W,ones(NN,1))*matmul(W,ones(NN,1)))+sum( matmul(W,ones(NN,1))))&
                    &+(gamma_1(6)-gamma_2(6))*(sum(matmul(ones(1,NN),W)*matmul(ones(1,NN),W))-sum( matmul(W,ones(NN,1)))) &
                    &+(gamma_1(7)-gamma_2(7))*trace( matmul(matmul(W,W),transpose(W)),NN) &
                    &+(gamma_1(8)-gamma_2(8))*trace( matmul(matmul(W,W),W),NN) &
                    &+0.5*gamma_2(9)*dot_product(Y3,Y3)-0.5*gamma_1(9)*dot_product(Y4,Y4) &
                    &+trace(matmul(psi_2,transpose(w_new)),NN)-trace(matmul(psi_1,transpose(w_new)),NN)  &
                    &+(gamma_2(3)-gamma_1(3))*trace(matmul(w_new,w_new),NN)&
                    &+(gamma_2(4)-gamma_1(4))*(sum(matmul(w_new,ones(NN,1))*matmul(w_new,ones(NN,1)))-sum(matmul(w_new,ones(NN,1)))) &
                    &+(gamma_2(5)-gamma_1(5))*(sum(matmul(w_new,ones(NN,1))*(matmul(w_new,ones(NN,1))*matmul(w_new,ones(NN,1))))&
                    &-2*sum(matmul(w_new,ones(NN,1))*matmul(w_new,ones(NN,1)))+sum(matmul(w_new,ones(NN,1))))&
                    &+(gamma_2(6)-gamma_1(6))*(sum(matmul(ones(1,NN),w_new)*matmul(ones(1,NN),w_new))-sum(matmul(w_new,ones(NN,1)))) &
                    &+(gamma_2(7)-gamma_1(7))*trace( matmul(matmul(w_new,w_new),transpose(w_new)),NN) &
                    &+(gamma_2(8)-gamma_1(8))*trace( matmul(matmul(w_new,w_new),w_new),NN)


                q_g1(g)=q_1
                pp_l(g)=(like_Y1-like_Y2)

                DEALLOCATE(C,W,Wt,Y,X,XX,ep_1,ep_2,psi,psi_1,psi_2,w_old,w_new,Wt_old,Wt_new,pp_w,h1,h2,Y1,Y2,Y1s,Y2s,Y3,Y3s,Y4,Y4s, &
                    & W12, W22, Wt12, Wt22, S22, S22_new, S22_old)

            END DO
            !$omp end parallel
            pp_g=sum(pp_l)+sum(q_g1)+logmvnpdf(gamma_1, G_0, 9)-logmvnpdf(gamma_2, G_0, 9)
            CALL random_number(rand)
            IF (log(rand)<=pp_g)  THEN
                gamma_T(:,t)=gamma_1
                lambda_T(t)=lambda_1
                acc_1=acc_1+1.0
            ELSE
                gamma_T(:,t)=gamma_T(:,t-1)
                lambda_T(t)=lambda_T(t-1)
            END IF
            acc_rate1(t)=acc_1/t



            !********************************************************!
            !******************  DMH STEP FOR BETA  *****************!
            !********************************************************!

            B_1=eye(2)*c_2
            call spotrf('U',2, B_1,2,info)
            errcode=vsrnggaussianmv(method,stream,1,beta_1,2,me,beta_0, B_1)
            beta_1=beta_1+beta_T(:,t-1)

            IF (t>2000) THEN
                B_1=cov(beta_T(:,1:t-1),2,t-1)*2.38**2/(2)
                call spotrf('U',2, B_1, 2, info)
                errcode=vsrnggaussianmv(method,stream,1, beta_2, 2, me, beta_0, B_1)
                beta_2=beta_2+beta_T(:,t-1)
                beta_1=0.95*beta_2+0.05*beta_1
            END IF

            pp_l=0.0

            Call OMP_SET_NUM_THREADS(Nthreads)
            !$omp parallel default(shared) private(g, NN, C, W,Wt,Y,X,XX, w1, w2, w3, w4, psi,    &
            !$omp & w12, w22, Wt12, Wt22, S22, ep_1, ep_2, i, j, r, W_old,W_new,Wt_old, Wt_new,      &
            !$omp &  q, pp_w, rand, INFO, like_Y1, like_Y2, h1, h2,  Y1, Y2,  Y1s, Y2s, Y3, Y4, Y3s, Y4s, &
            !$omp & ithread, iblock, ibegin, iend)

            ithread = OMP_GET_THREAD_NUM()
            iblock = (GG+Nthreads-1)/Nthreads
            ibegin = ithread*iblock+1
            iend = min((ithread+1)*iblock,GG)
            DO g = ibegin,iend

                NN=N(g)
                ALLOCATE(Y(NN));
                ALLOCATE(C(NN,NN));        ALLOCATE(psi(NN,NN));
                ALLOCATE(W(NN,NN));        ALLOCATE(Wt(NN,NN))
                ALLOCATE(X(NN));           ALLOCATE(XX(NN,2))
                ALLOCATE(w_old(NN,NN));    ALLOCATE(w_new(NN,NN));
                ALLOCATE(Wt_old(NN,NN));   ALLOCATE(Wt_new(NN,NN));
                ALLOCATE(ep_1(NN));        ALLOCATE(ep_2(NN))
                ALLOCATE(h1(NN));          ALLOCATE(h2(NN))
                ALLOCATE(pp_w(NN,NN))
                ALLOCATE(Y1(NN));          ALLOCATE(Y2(NN))
                ALLOCATE(Y1s(NN));         ALLOCATE(Y2s(NN))
                ALLOCATE(Y3(NN));          ALLOCATE(Y4(NN))
                ALLOCATE(Y3s(NN));         ALLOCATE(Y4s(NN))

                W=W_raw(1:NN,1:NN,g)
                Y=Y_latent(1:NN,g)
                X=X_raw(1:NN,g)
                Wt=transpose(W)

                XX(:,1)=X
                XX(:,2)=matmul(W,X)

                psi=0.0
                DO i=1, NN
                    DO j=1, NN
                        IF (j/=i) THEN
                            psi(i,j)=gamma_T(1,t)+gamma_T(2,t)*C(i,j)
                        END IF
                    END DO
                END DO


                ALLOCATE(W12(censor(g),NN-censor(g)))
                ALLOCATE(W22(NN-censor(g),NN-censor(g)))
                ALLOCATE(Wt12(censor(g),NN-censor(g)))
                ALLOCATE(Wt22(NN-censor(g),NN-censor(g)))
                ALLOCATE(S22(NN-censor(g),NN-censor(g)))

                W12=W(1:censor(g),censor(g)+1:NN)
                W22=W(censor(g)+1:NN,censor(g)+1:NN)
                Wt12=Wt(1:censor(g),censor(g)+1:NN)
                Wt22=Wt(censor(g)+1:NN,censor(g)+1:NN)
                S22=eye(NN-censor(g))-lambda_T(t)*(W22)

                !! Define Epsilon !!
                ep_1(1:censor(g))=Y(1:censor(g))-( lambda_T(t)*matmul(W12,Y(censor(g)+1:NN))+matmul(XX(1:censor(g),:),beta_1)+alpha_T(g,t-1))
                ep_1(censor(g)+1:NN)=matmul(S22,Y(censor(g)+1:NN))-matmul(XX(censor(g)+1:NN,:),beta_1)-alpha_T(g,t-1)

                ep_2(1:censor(g))=Y(1:censor(g))-( lambda_T(t)*matmul(W12,Y(censor(g)+1:NN))+matmul(XX(1:censor(g),:),beta_T(:,t-1))+alpha_T(g,t-1))
                ep_2(censor(g)+1:NN)=matmul(S22,Y(censor(g)+1:NN))-matmul(XX(censor(g)+1:NN,:),beta_T(:,t-1))-alpha_T(g,t-1)


                W_new=W
                W_old=W
                Wt_new=Wt
                Wt_old=Wt

                XX(:,2)=matmul(w_old,X)
                h1=ep_1+matmul(XX,beta_1)+alpha_T(g,t-1)
                CALL contraction(NN,h1,W_old,lambda_T(t),Y2)

                DO r=1, RR
                    DO i=1, NN
                        DO j=1, NN
                            IF (j /= i) THEN
                                w_new(i,j)=1-w_new(i,j)
                                XX(:,2)=matmul(w_new,X)
                                h1=ep_1+matmul(XX,beta_1)+alpha_T(g,t-1)
                                CALL contraction(NN,h1,W_new,lambda_T(t),Y1)

                                w1=dot_product(w_new(i,:),w_new(j,:))
                                w2=dot_product(w_new(:,i),w_new(:,j))
                                w3=dot_product(w_new(i,:),w_new(:,j))
                                w4=dot_product(w_new(j,:),w_new(:,i))

                                q=psi(i,j)+2*gamma_T(3,t)*w_new(j,i)+(2*gamma_T(4,t)-gamma_T(5,t))*(sum(w_new(i,:))-w_new(i,j)) &
                                    &+3*gamma_T(5,t)*(sum(w_new(i,:))-w_new(i,j))**2 &
                                    &+2*gamma_T(6,t)*(sum(w_new(:,j))-w_new(i,j)) +gamma_T(7,t)*(w1+w2+w3)+gamma_T(8,t)*3*w4

                                pp_w(i,j)=q*(-1)**(1-w_new(i,j))+0.5*gamma_T(9,t)*(dot_product(Y1,Y1)-dot_product(Y2,Y2))

                                CALL random_number(rand)
                                IF (log(rand)<=pp_w(i,j)) then
                                    W_old(i,j)=W_new(i,j)
                                    Y2=Y1
                                END IF
                                W_new(i,j)=W_old(i,j)
                                XX(:,2)=matmul(w_new,X)
                            END IF
                        END DO
                    END DO
                END DO


                like_Y1=-dot_product(ep_1, ep_1)/(2*sige_T(g,t-1))
                like_Y2=-dot_product(ep_2, ep_2)/(2*sige_T(g,t-1))



                XX(:,2)=matmul(w_new,X)
                h1=ep_1+matmul(XX,beta_1       )+alpha_T(g,t-1)
                h2=ep_2+matmul(XX,beta_T(:,t-1))+alpha_T(g,t-1)
                CALL contraction(NN,h2,W_new,lambda_T(t),Y3)
                CALL contraction(NN,h1,W_new,lambda_T(t),Y4)


                pp_l(g)=(like_Y1-like_Y2)+0.5*gamma_T(9,t)*(dot_product(Y3,Y3)-dot_product(Y4,Y4))


                DEALLOCATE(C,W,Wt,Y,X,XX,psi,w_old,w_new,Wt_old,Wt_new,ep_1,ep_2,h1,h2,pp_w,Y1,Y2,Y1s,Y2s,Y3,Y4,Y3s,Y4s,w12,w22,S22, wt12, wt22)


            END DO
            !$omp end parallel

            pp_g=sum(pp_l)+logmvnpdf(beta_1, B_0, 2)-logmvnpdf(beta_T(:,t-1), B_0, 2)
            CALL random_number(rand)
            IF (log(rand)<=pp_g)  THEN
                beta_T(:,t)=beta_1
                acc_2=acc_2+1.0
            ELSE
                beta_T(:,t)=beta_T(:,t-1)
            END IF
            acc_rate2(t)=acc_2/t


            !******************GIBBS STEP FOR SIGMA_E  *****************!

            DO g=1,GG
                NN=N(g)
                ALLOCATE(W(NN,NN));
                ALLOCATE(Wt(NN,NN))
                ALLOCATE(Y(NN))
                ALLOCATE(X(NN));            ALLOCATE(XX(NN,2))
                ALLOCATE(ep(NN))

                Y=Y_latent(1:NN,g)
                W=W_raw(1:NN,1:NN,g)
                Wt=transpose(W)
                X=X_raw(1:NN,g)

                XX(:,1)=X
                XX(:,2)=matmul(W,X)

                ALLOCATE(W12(censor(g),NN-censor(g)))
                ALLOCATE(W22(NN-censor(g),NN-censor(g)))
                ALLOCATE(Wt12(censor(g),NN-censor(g)))
                ALLOCATE(Wt22(NN-censor(g),NN-censor(g)))
                ALLOCATE(S22(NN-censor(g),NN-censor(g)))

                W12=W(1:censor(g),censor(g)+1:NN)
                W22=W(censor(g)+1:NN,censor(g)+1:NN)
                Wt12=Wt(1:censor(g),censor(g)+1:NN)
                Wt22=Wt(censor(g)+1:NN,censor(g)+1:NN)
                S22=eye(NN-censor(g))-lambda_T(t)*(W22)


                !! Define Epsilon !!
                ep(1:censor(g))=Y(1:censor(g))-( lambda_T(t)*matmul(W12,Y(censor(g)+1:NN))+matmul(XX(1:censor(g),:),beta_T(:,t))+alpha_T(g,t-1))
                ep(censor(g)+1:NN)=matmul(S22,Y(censor(g)+1:NN))-matmul(XX(censor(g)+1:NN,:),beta_T(:,t))-alpha_T(g,t-1)


                sige_T(g,t)=(dot_product(ep,ep)+eta_0)/random_chisq(NN+rho_0, first)

                DEALLOCATE(W12,Wt12,W22,Wt22,S22,W,Wt,X,XX,Y,ep)
            END DO


            !!*************GIBBS STEP FOR ALPHA_G*********************!


            Call OMP_SET_NUM_THREADS(Nthreads)
            !$omp parallel default(shared) private(g, NN, C, W,Wt,Y,X,XX, w1, w2, w3, w4, psi, alpha_1, alpha_2,   &
            !$omp & w12, w22, Wt12, Wt22, S22, ep_1, ep_2, i, j, r, W_old,W_new,Wt_old, Wt_new,      &
            !$omp &  q, pp_w, pp, rand, INFO, like_Y1, like_Y2, h1, h2,  Y1, Y2,  Y1s, Y2s, Y3, Y4, Y3s, Y4s, &
            !$omp & ithread, iblock, ibegin, iend)

            ithread = OMP_GET_THREAD_NUM()
            iblock = (GG+Nthreads-1)/Nthreads
            ibegin = ithread*iblock+1
            iend = min((ithread+1)*iblock,GG)

            DO g = ibegin,iend

                NN=N(g)
                ALLOCATE(Y(NN));
                ALLOCATE(C(NN,NN));        ALLOCATE(psi(NN,NN));
                ALLOCATE(W(NN,NN));        ALLOCATE(Wt(NN,NN))
                ALLOCATE(X(NN));           ALLOCATE(XX(NN,2))
                ALLOCATE(w_old(NN,NN));    ALLOCATE(w_new(NN,NN));
                ALLOCATE(Wt_old(NN,NN));   ALLOCATE(Wt_new(NN,NN));
                ALLOCATE(ep_1(NN));        ALLOCATE(ep_2(NN))
                ALLOCATE(h1(NN));          ALLOCATE(h2(NN))
                ALLOCATE(pp_w(NN,NN))
                ALLOCATE(Y1(NN));          ALLOCATE(Y2(NN))
                ALLOCATE(Y1s(NN));         ALLOCATE(Y2s(NN))
                ALLOCATE(Y3(NN));          ALLOCATE(Y4(NN))
                ALLOCATE(Y3s(NN));         ALLOCATE(Y4s(NN))


                alpha_1=alpha_T(g,t-1)+random_normal()*sqrt(c_3(g))
                IF (t>100) THEN
                    alpha_2=random_normal()*sqrt(var(alpha_T(g,1:t-1),t-1))*2.38+alpha_T(g,t-1)
                    alpha_1=0.95*alpha_2+0.05*alpha_1
                END IF

                W=W_raw(1:NN,1:NN,g)
                Wt=transpose(W)
                Y=Y_latent(1:NN,g)
                X=X_raw(1:NN,g)

                XX(:,1)=X
                XX(:,2)=matmul(W,X)

                psi=0.0
                DO i=1, NN
                    DO j=1, NN
                        IF (j/=i) THEN
                            psi(i,j)=gamma_T(1,t)+gamma_T(2,t)*C(i,j)
                        END IF
                    END DO
                END DO


                ALLOCATE(W12(censor(g),NN-censor(g)))
                ALLOCATE(W22(NN-censor(g),NN-censor(g)))
                ALLOCATE(Wt12(censor(g),NN-censor(g)))
                ALLOCATE(Wt22(NN-censor(g),NN-censor(g)))
                ALLOCATE(S22(NN-censor(g),NN-censor(g)))

                W12=W(1:censor(g),censor(g)+1:NN)
                W22=W(censor(g)+1:NN,censor(g)+1:NN)
                Wt12=Wt(1:censor(g),censor(g)+1:NN)
                Wt22=Wt(censor(g)+1:NN,censor(g)+1:NN)
                S22=eye(NN-censor(g))-lambda_T(t)*(W22)

                !! Define Epsilon !!
                ep_1(1:censor(g))=Y(1:censor(g))-(lambda_T(t)*matmul(W12,Y(censor(g)+1:NN))+matmul(XX(1:censor(g),:),beta_T(:,t))+alpha_1)
                ep_1(censor(g)+1:NN)=matmul(S22,Y(censor(g)+1:NN))-matmul(XX(censor(g)+1:NN,:),beta_T(:,t))-alpha_1

                ep_2(1:censor(g))=Y(1:censor(g))-(lambda_T(t)*matmul(W12,Y(censor(g)+1:NN))+matmul(XX(1:censor(g),:),beta_T(:,t))+alpha_T(g,t-1))
                ep_2(censor(g)+1:NN)=matmul(S22,Y(censor(g)+1:NN))-matmul(XX(censor(g)+1:NN,:),beta_T(:,t))-alpha_T(g,t-1)

                W_new=W
                W_old=W
                Wt_new=Wt
                Wt_old=Wt

                XX(:,2)=matmul(w_old,X)
                h1=ep_1+matmul(XX,beta_T(:,t))+alpha_1
                CALL contraction(NN,h1,W_old,lambda_T(t),Y2)

                DO r=1, RR
                    DO i=1, NN
                        DO j=1, NN
                            IF (j /= i) THEN
                                w_new(i,j)=1-w_new(i,j)
                                XX(:,2)=matmul(w_new,X)
                                h1=ep_1+matmul(XX,beta_T(:,t))+alpha_1
                                CALL contraction(NN,h1,W_new,lambda_T(t),Y1)

                                w1=dot_product(w_new(i,:),w_new(j,:))
                                w2=dot_product(w_new(:,i),w_new(:,j))
                                w3=dot_product(w_new(i,:),w_new(:,j))
                                w4=dot_product(w_new(j,:),w_new(:,i))

                                q=psi(i,j)+2*gamma_T(3,t)*w_new(j,i)+(2*gamma_T(4,t)-gamma_T(5,t))*(sum(w_new(i,:))-w_new(i,j)) &
                                    &+3*gamma_T(5,t)*(sum(w_new(i,:))-w_new(i,j))**2 &
                                    &+2*gamma_T(6,t)*(sum(w_new(:,j))-w_new(i,j)) +gamma_T(7,t)*(w1+w2+w3)+gamma_T(8,t)*3*w4

                                pp_w(i,j)=q*(-1)**(1-w_new(i,j))+0.5*gamma_T(9,t)*(dot_product(Y1,Y1)-dot_product(Y2,Y2))

                                CALL random_number(rand)
                                IF (log(rand)<=pp_w(i,j)) then
                                    W_old(i,j)=W_new(i,j)
                                    Y2=Y1
                                END IF
                                W_new(i,j)=W_old(i,j)
                                XX(:,2)=matmul(w_new,X)
                            END IF
                        END DO
                    END DO
                END DO


                like_Y1=-dot_product(ep_1, ep_1)/(2*sige_T(g,t))
                like_Y2=-dot_product(ep_2, ep_2)/(2*sige_T(g,t))

                XX(:,2)=matmul(w_new,X)
                h1=ep_1+matmul(XX,beta_T(:,t))+alpha_1
                h2=ep_2+matmul(XX,beta_T(:,t))+alpha_T(g,t-1)
                CALL contraction(NN,h2,W_new,lambda_T(t),Y3)
                CALL contraction(NN,h1,W_new,lambda_T(t),Y4)


                pp=(like_Y1-like_Y2)+0.5*gamma_T(9,t)*(dot_product(Y3,Y3)-dot_product(Y4,Y4))+((alpha_T(g,t-1)**2-alpha_1**2)/(2*A_0))

                CALL random_number(rand)
                IF (log(rand)<=pp)  THEN
                    alpha_T(g,t)=alpha_1
                    acc_3(g)=acc_3(g)+1.0
                ELSE
                    alpha_T(g,t)=alpha_T(g,t-1)
                END IF
                acc_rate3(g,t)=acc_3(g)/t
                !IF((acc_rate3(g,t)<=0.4) .and. (c_3(g)>=1e-6)) c_3(g)=c_3(g)/1.01
                !IF((acc_rate3(g,t)>=0.6) .and. (c_3(g)<=2)) c_3(g)=c_3(g)*1.01

                DEALLOCATE(W,Wt,Y,X,XX,C,psi,w_old,w_new,Wt_old,Wt_new,ep_1,ep_2,h1,h2,pp_w,Y1,Y1s,Y2,Y2s,Y3,Y3s,Y4,Y4s,w12,w22,S22, wt12, wt22)

            END DO
            !$omp end parallel

            call system_clock ( clock_stop, clock_rate, clock_max )
            t_round=t/10
            IF (t-t_round*10==0) THEN
                WRITE(*,*)   'T=', t
                WRITE(*,10) 'Time:', real ( clock_stop - clock_start ) / real ( clock_rate )
                WRITE(*,'(a6, 9f7.4)') 'GAMMA:', gamma_T(:,t)
                WRITE(*,10) 'LAMBDA:', lambda_T(t)
                WRITE(*,10) 'BETA:',   beta_T(:,t)
                WRITE(*,10) 'SIGE:',   sige_T(1,t), sige_T(11,t)
                WRITE(*,10) 'acc_1:',  acc_rate1(t)
                WRITE(*,10) 'acc_2:',  acc_rate2(t)
                WRITE(*,10) 'acc_3:',  acc_rate3(1,t)
            END IF
            WRITE(1000,'(14f10.4)')  gamma_T(:,t)
            WRITE(2000,'(f10.4)')    lambda_T(t)
            WRITE(3000,'(30f10.4)')  beta_T(:,t)
            WRITE(4000,'(47f10.4)')  sige_T(:,t)
            WRITE(5000,'(47f10.4)')  alpha_T(:,t)

        END DO
    END DO

10  format(a6,5f7.4)
30  format(a6,8f7.4)
40  format(a6,6f10.4)
50  format(a6,5f10.4)
60  format(a6,f10.4)
70  format(30f10.4)


    end program estimation

    SUBROUTINE contraction(NN,ep,W,lambda,Y)
    IMPLICIT NONE
    INTEGER:: NN
    REAL , INTENT(IN)  :: lambda, ep(:),  W(:,:)
    REAL, INTENT(OUT)  :: Y(:)

    ! local variable !
    REAL, ALLOCATABLE :: Y1(:), Y2(:), temp_Y(:,:)
    REAL :: diff
    INTEGER:: count

    ALLOCATE(Y1(NN))
    ALLOCATE(Y2(NN))
    ALLOCATE(temp_Y(NN,2))

    Y1=ep
    diff=1.0
    count=0
    Do
        temp_Y(:,1)=lambda*matmul(W,Y1)+ep
        temp_Y(:,2)=0.0
        Y2=maxval(temp_Y,dim=2)
        diff=maxval(abs(Y2-Y1))
        if (diff<=1e-6) exit
        Y1=Y2
        count=count+1
        if (count>=200) then
            !print*, 'convergence fail'
            exit
        endif
    End do
    Y=Y2

    DEALLOCATE(Y1)
    DEALLOCATE(Y2)
    DEALLOCATE(temp_Y)
    END SUBROUTINE contraction