    include 'mkl_vsl.f90'

    program DGP

    USE random
    USE TOOL
    USE lapack95
    USE MKL_VSL_TYPE
    USE MKL_VSL
    USE omp_lib
    IMPLICIT NONE


    !******************* GENERAL VARIABLES ****************!
    CHARACTER(100)           :: path1='D:/research/2020_exponential_network_formation/QE/simulation/continuous/'
    CHARACTER(100)           :: path2='/data/data_30_30/'
    CHARACTER(40)            :: numchr1, numchr2
    INTEGER, PARAMETER       :: NN=30, GG=30, RR=2000, TT=100, Nthreads=6
    REAL, ALLOCATABLE        :: C(:,:), W(:,:), Wt(:,:), W_new(:,:), Wt_new(:,:), X(:), Y(:), E(:), Z(:), zero(:), C1(:), Iden(:,:), IPIV(:)
    REAL, ALLOCATABLE        :: cov_x(:,:), cov_e(:,:), cov_z(:,:), psi(:,:), XX(:,:), ZZ(:,:), h(:)
    REAL, ALLOCATABLE        :: S(:,:), S1(:,:), S2(:,:),S1L(:,:), S2L(:,:), S_new(:,:)
    REAL, ALLOCATABLE        :: Y1(:), Y2(:), Y1s(:), Y2s(:), Y_new(:)
    REAL                     :: gamma(9), beta(2), rho(2), lambda, delta, siga, sige, alpha, kappa, mu_z, rand(1)
    REAL(kind=8)             :: w1, w2, w3, w4, q, pp_w
    INTEGER                  :: clock_max, clock_rate, clock_start, clock_stop, ithread, iblock,ibegin,iend
    INTEGER                  :: brng, method_gaussian, method_uniform, me, i, j, g, t, flag, ier, r, seedsize, seed(2), errorflag, INFO, seed_mkl
    LOGICAL, PARAMETER       :: first=.true.
    TYPE (VSL_STREAM_STATE)  :: stream_gaussian, stream_uniform
    INTEGER                  :: errcode
    !*******************************************************!

    seedsize=2
    seed(1)=2.1103e+005
    seed(2)=2.1103e+005
    CALL RANDOM_SEED
    CALL RANDOM_SEED(SIZE=seedsize)
    CALL RANDOM_SEED(PUT=seed)

    seed_mkl=12345
    brng=VSL_BRNG_MT2203
    method_uniform=VSL_RNG_METHOD_UNIFORM_STD
    method_gaussian=VSL_RNG_METHOD_GAUSSIAN_ICDF
    me=VSL_MATRIX_STORAGE_FULL

    errcode=vslnewstream(stream_gaussian,brng,seed_mkl)
    errcode=vslnewstream(stream_uniform,brng,seed_mkl)

    !******* Parameters **********!
    gamma=(/-0.5, 0.3, 0.3, 0.20, -0.10, 0.04, 0.30, -0.20,  1.0 /)
    beta=(/ 0.3, 0.1 /)
    rho=(/ 0.3, 0.1  /)
    delta=0.30
    sige=0.5
    lambda=0.05
    siga=0.3
    !***************************!


    
    call system_clock ( clock_start, clock_rate, clock_max )



    DO t=1, TT

        print*, 'repetition', t
        WRITE(numchr1,*) t

        Call OMP_SET_NUM_THREADS(Nthreads)
        !$omp parallel default(shared) private(g,C1,C,i,j,Iden,INFO,IPIV,X,XX,cov_x,cov_z,cov_e,Z,ZZ, E, numchr2, &
        !$omp & zero,alpha,psi,h,W,W_new,r,S1,S2,Y1,Y2,Y1s,Y2s,w1,w2,w3,w4,q, pp_w, rand, S_new, Y_new, errcode,  &
        !$omp & ithread, iblock, ibegin, iend)


        ithread = OMP_GET_THREAD_NUM()
        iblock = (GG+Nthreads-1)/Nthreads
        ibegin = ithread*iblock+1
        iend = min((ithread+1)*iblock,GG)

        DO g = ibegin,iend
            print*, g

            ALLOCATE(C(NN,NN));      ALLOCATE(C1(NN))
            ALLOCATE(W(NN,NN));
            ALLOCATE(W_new(NN,NN));
            ALLOCATE(cov_x(NN,NN));  ALLOCATE(cov_e(NN,NN))
            ALLOCATE(cov_z(NN,NN))
            ALLOCATE(X(NN));         ALLOCATE(XX(NN,2))
            ALLOCATE(Y1(NN));        ALLOCATE(Y2(NN));
            ALLOCATE(Y1s(NN));       ALLOCATE(Y2s(NN));
            ALLOCATE(zero(NN));      ALLOCATE(E(NN));
            ALLOCATE(Iden(NN,NN));   ALLOCATE(IPIV(NN))
            ALLOCATE(psi(NN,NN));
            ALLOCATE(h(NN));         ALLOCATE(S_new(NN,NN))
            ALLOCATE(Y_new(NN));
            ALLOCATE(Z(NN));         ALLOCATE(ZZ(NN,2))
            ALLOCATE(S1(NN,NN));     ALLOCATE(S2(NN,NN))


            DO
                CALL random_number(C1)

                C=0.0
                DO i=1,NN
                    DO j=1,NN
                        if((C1(i)>=0.7) .and. (C1(j)>=0.7)) C(i,j)=1.0
                        if((C1(i)<=0.3) .and. (C1(j)<=0.3)) C(i,j)=1.0
                    END DO
                    C(i,i)=0.0
                END DO

                zero=0.0
                Iden=eye(NN)

                cov_x=Iden
                call spotrf('U', NN, cov_x, NN,INFO)
                errcode=vsrnggaussianmv(method_gaussian,stream_gaussian, 1, X, NN, me, zero, cov_x)

                XX(:,1)=X

                cov_z=Iden
                call spotrf('U', NN, cov_z, NN,INFO)
                errcode=vsrnggaussianmv(method_gaussian,stream_gaussian, 1, Z, NN, me, zero, cov_z)
                ZZ(:,1)=Z


                cov_e=Iden*sige
                call spotrf('U', NN, cov_e, NN,INFO)
                errcode=vsrnggaussianmv(method_gaussian,stream_gaussian, 1, E, NN, me, zero, cov_e)

                alpha=3.0+random_normal()


                psi=0.0
                DO i=1, NN
                    DO j=1, NN
                        IF (j /= i) THEN
                            psi(i,j)=gamma(1)+gamma(2)*C(i,j)-gamma(9)*abs(Z(i)-Z(j))
                        END IF
                    END DO
                END DO


                W=0.0
                S2=eye(NN)-(W)*lambda
                XX(:,2)=matmul(W,X)
                ZZ(:,2)=matmul(W,Z)
                Y2=matmul(XX,beta)+matmul(ZZ,rho)+alpha+E
                CALL SGESV(NN,1,S2,NN,IPIV,Y2,NN,INFO)

                W_new=W

                DO r=1, RR
                    errcode=vsrnguniform( method_uniform, stream_uniform, 1, rand, 0.0, 1.0)

                    DO i=1, NN
                        DO j=1, NN
                            IF (j /= i) THEN
                                W_new(i,j)=1.0-W_new(i,j)
                                S1=eye(NN)-W_new*lambda
                                XX(:,2)=matmul(W_new,X)
                                ZZ(:,2)=matmul(W_new,Z)
                                Y1=matmul(XX,beta)+matmul(ZZ,rho)+alpha+E
                                CALL SGESV(NN,1,S1,NN,IPIV,Y1,NN,INFO)

                                w1=dot_product(W_new(i,:),W_new(j,:))
                                w2=dot_product(W_new(:,i),W_new(:,j))
                                w3=dot_product(W_new(i,:),W_new(:,j))
                                w4=dot_product(W_new(j,:),W_new(:,i))

                                q=psi(i,j)+2*gamma(3)*W_new(j,i)+(2*gamma(4)-gamma(5))*(sum(W_new(i,:))-W_new(i,j)) &
                                    &+3*gamma(5)*(sum(W_new(i,:))-W_new(i,j))**2 &
                                    &+2*gamma(6)*(sum(W_new(:,j))-W_new(i,j))+gamma(7)*(w1+w2+w3)+gamma(8)*3*w4

                                pp_w=q*(-1)**(1-w_new(i,j))+0.5*delta*(dot_product(Y1,Y1)-dot_product(Y2,Y2))

                                !CALL random_number(rand)
                                errcode=vsrnguniform( method_uniform, stream_uniform, 1, rand, 0.0, 1.0)
                                IF (log(rand(1))<=pp_w) then
                                    W(i,j) =W_new(i,j)
                                    Y2=Y1
                                END IF
                                W_new(i,j)=W(i,j)
                                XX(:,2)=matmul(W_new,X)
                                ZZ(:,2)=matmul(W_new,Z)
                            END IF
                        END DO
                    END DO
                END DO
                IF ((maxval(sum(W, dim=2))<=25.0) .and. (maxval(sum(W, dim=1))<=25.0)) THEN
                    EXIT
                ELSE
                    print*, 'degenerate'
                END IF
            END DO


            print*, sum(sum(W, dim=2))/NN

            S_new=eye(NN)-W*lambda
            XX(:,2)=matmul(W,X)
            ZZ(:,2)=matmul(W,Z)
            Y_new=matmul(XX,beta)+matmul(ZZ,rho)+alpha+E
            CALL SGESV(NN,1,S_new,NN,IPIV,Y_new,NN,INFO)


            WRITE(numchr2,*) g

            OPEN (10000+g, file = trim(adjustl(path1))  // trim(adjustl(path2)) // 'R' // trim(adjustl(numchr1)) // '/C/C' // trim(adjustl(numchr2)) // '.txt', ACTION='READWRITE', status='unknown')
            OPEN (20000+g, file = trim(adjustl(path1))  // trim(adjustl(path2)) // 'R' // trim(adjustl(numchr1)) // '/W/W' // trim(adjustl(numchr2)) // '.txt', ACTION='READWRITE', status='unknown')
            OPEN (30000+g, file = trim(adjustl(path1))  // trim(adjustl(path2)) // 'R' // trim(adjustl(numchr1)) // '/X/X' // trim(adjustl(numchr2)) // '.txt', ACTION='READWRITE', status='unknown')
            OPEN (40000+g, file = trim(adjustl(path1))  // trim(adjustl(path2)) // 'R' // trim(adjustl(numchr1)) // '/Y/Y' // trim(adjustl(numchr2)) // '.txt', ACTION='READWRITE', status='unknown')
            OPEN (50000+g, file = trim(adjustl(path1))  // trim(adjustl(path2)) // 'R' // trim(adjustl(numchr1)) // '/alpha/alpha' // trim(adjustl(numchr2)) // '.txt', ACTION='READWRITE', status='unknown')
            OPEN (60000+g, file = trim(adjustl(path1))  // trim(adjustl(path2)) // 'R' // trim(adjustl(numchr1)) // '/Z/Z' // trim(adjustl(numchr2)) // '.txt', ACTION='READWRITE', status='unknown')


            IF(NN==25) THEN
                write(10000+g,'(25f6.2)') ((C(i,j),j=1,NN),i=1,NN)
                write(20000+g,'(25f6.2)') ((W(i,j),j=1,NN),i=1,NN)
            ELSEIF (NN==30) THEN
                write(10000+g,'(30f6.2)') ((C(i,j),j=1,NN),i=1,NN)
                write(20000+g,'(30f6.2)') ((W(i,j),j=1,NN),i=1,NN)
            ELSEIF (NN==40) THEN
                write(10000+g,'(40f6.2)') ((C(i,j),j=1,NN),i=1,NN)
                write(20000+g,'(40f6.2)') ((W(i,j),j=1,NN),i=1,NN)
            END IF


            write(30000+g,300) (X(i),i=1,NN)
            write(40000+g,400) (Y_new(i), i=1,NN)
            write(50000+g,400) (alpha)
            write(60000+g,300) (Z(i),i=1,NN)


            close(10000+g)
            close(20000+g)
            close(30000+g)
            close(40000+g)
            close(50000+g)
            close(60000+g)

            DEALLOCATE(C,C1,W,W_new,cov_x,cov_z,cov_e,X,XX,Y1,Y2,Y1s,Y2s,zero,E,Iden,IPIV,psi,h,S_new,Y_new,Z,ZZ,S1,S2)


        end do
        !$omp end parallel

        call system_clock ( clock_stop, clock_rate, clock_max )
        print*, 'Time:', real ( clock_stop - clock_start ) / real ( clock_rate )

    END DO

    
    !***** Deinitialize *****

    errcode=vsldeletestream( stream_gaussian )
    errcode=vsldeletestream( stream_uniform )
    pause


300 format (1F10.4)
400 format (1F10.4)

    end program DGP

